###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import json
from urllib import request, parse

from starthinker.util.project import project
from simple_salesforce import Salesforce


def authenticate(domain, client_id, client_secret, username, password):

  data = parse.urlencode({
      'grant_type': 'password',
      'client_id': client_id,
      'client_secret': client_secret,
      'username': username,
      'password': password
  }).encode()

  req = request.Request('https://%s/services/oauth2/token' % domain, data=data)
  creds = json.loads(request.urlopen(req).read())
  return Salesforce(
      instance_url=creds['instance_url'], session_id=creds['access_token'])


def query(service, query, header=False):
  for record in service.query_all(query)['records']:
    if header:
      yield [column for column in record.keys() if column != 'attributes']
      header = False
    yield [record[column] for column in record.keys() if column != 'attributes']



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from django.core.management.base import BaseCommand, CommandError
from django.conf import settings

from starthinker_ui.recipe.scripts import Script
from starthinker_ui.recipe.colab import script_to_colab


class Command(BaseCommand):
  help = 'Generate Templates For Colab'

  def handle(self, *args, **kwargs):

    with open('%s/tutorials/deploy_colab.md' % settings.UI_ROOT,
              'w') as readme_file:
      readme_file.write('# Using Scripts As A Colab Notebook\n')
      readme_file.write('\n')
      readme_file.write(
          'All StarThinker recipes and solutions can be run from [Google Collaboratory](https://colab.research.google.com/github/google/starthinker/blob/master). '
      )
      readme_file.write(
          'Also visit the [Solution Gallery](google.github.io/starthinker/) or click a link below to deploy a notebook.\n'
      )
      readme_file.write('\n')
      readme_file.write('## List Of Notebooks\n')

      for script in Script.get_scripts():
        if script.get_open_source():
          readme_file.write('* [%s](%s) - %s\n' %
                            (script.get_name(), script.get_link_colab(),
                             script.get_description()))

      readme_file.write('---\n')
      readme_file.write(
          '&copy; 2019 Google LLC - Apache License, Version 2.0\n')

    for script in Script.get_scripts():
      if script.get_open_source():
        print('Writing: %s/colabs/%s.ipynb' %
              (settings.UI_ROOT, script.get_tag()))
        with open('%s/colabs/%s.ipynb' % (settings.UI_ROOT, script.get_tag()),
                  'w') as colab_file:
          colab_file.write(
              script_to_colab(
                  script.get_name(),
                  script.get_description(),
                  script.get_instructions(),
                  script.get_tasks(),
              ))



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################
"""AirFlow connector for StarThinker, builds airflow workflow from json recipe.

Connector between AirFlow and StarThinker that turns all tasks into
PythonOperators.  In future, as AirFlow matures and becomes more reliable,
some handlers will migrate from StarThinker tasks to native AirFlow calls.

Any StarThinker task immediately becomes available as an Airflow endpoint using
this code. The DAG ID will be set to the the path of the JSON template with non
alphanumeric characters set to '.', to mimic python dot notation.

"""


from datetime import timedelta, datetime, date
from importlib import import_module
from random import randint

from airflow import DAG
from airflow.hooks.base_hook import BaseHook
from airflow.operators.python_operator import PythonOperator
from airflow.operators.bash_operator import BashOperator

from starthinker.script.parse import json_set_fields

CONNECTION_SERVICE = "starthinker_service"
CONNECTION_USER = "starthinker_user"


class DAG_Factory():
  """A class factory that generates AirFlow Dag definitions from StarThinker JSON recipes.

  Called from an AirFlow Python Script, generates a single Dag mapped to each
  StarTHinker task handler.
  Loads a StartThinker JSON recipe and converts  it into an AirFlow DAG with
  Python operators.
  Each task int the JSON is converted to a PythonOperator, with the task passed
  in as a dict.
  The uuid inside the JSON recipe is used as the DAG ID.

  Sample calling wrapper, save this in your AirFlow dag directory:

    ```
    from starthinker.airflow.factory import DAG_Factory
    dag = DAG_Factory([StarThinker recipe JSON]).generate()
    ```

  """

  def __init__(self, _dag_name, _script, _script_parameters=None):
    self.dag_name = _dag_name
    self.recipe = _script
    self.dag = None

    if _script_parameters:
      json_set_fields(self.recipe, _script_parameters)

    self.recipe.setdefault('setup', {}).setdefault('auth', {})

    # If user credentials given in recipe, skip load from connection
    if self.recipe['setup']['auth'].get('user'):
      print('Loaded User Credentials From: RECIPE JSON')

    # If not given in recipe, try "user" auth information from connection
    else:
      try:
        user_connection_extra = BaseHook.get_connection(CONNECTION_USER).extra_dejson
        if user_connection_extra['extra__google_cloud_platform__key_path']:
          self.recipe['setup']['auth']['user'] = user_connection_extra[
            'extra__google_cloud_platform__key_path']
          print('Loaded User Credentials From: %s, Keyfile Path' % CONNECTION_USER)
        elif user_connection_extra['extra__google_cloud_platform__keyfile_dict']:
          self.recipe['setup']['auth']['user'] = user_connection_extra[
            'extra__google_cloud_platform__keyfile_dict']
          print('Loaded User Credentials From: %s, Keyfile JSON ' % CONNECTION_USER)
      except Exception as e:
        pass

    # If service credentials given in recipe, skip load from connection
    if self.recipe['setup']['auth'].get('service'):
      print('Loaded Service Credentials From RECIPE JSON')
    # If not given in recipe, try "service" auth information from connection
    else:
      try:
        service_connection_extra = BaseHook.get_connection(CONNECTION_SERVICE).extra_dejson
        if service_connection_extra['extra__google_cloud_platform__key_path']:
          self.recipe['setup']['auth']['service'] = service_connection_extra[
            'extra__google_cloud_platform__key_path']
          print('Loaded Service Credentials From: %s, Keyfile Path' % CONNECTION_SERVICE)
        elif service_connection_extra[
          'extra__google_cloud_platform__keyfile_dict']:
          self.recipe['setup']['auth']['service'] = service_connection_extra[
            'extra__google_cloud_platform__keyfile_dict']
          print('Loaded Service Credentials From: %s, Keyfile JSON' % CONNECTION_SERVICE)
      except Exception as e:
        pass

    # If project id given in recipe, skip load from connection
    if self.recipe['setup'].get('id'):
      print('Loaded Project ID From: RECIPE JSON')
    # If not given in recipe, try project id fetch from connections
    else:
      # check user
      try:
        user_connection_extra = BaseHook.get_connection(CONNECTION_USER).extra_dejson
        self.recipe['setup']['id'] = user_connection_extra.get('extra__google_cloud_platform__project')
      except: pass

      if self.recipe['setup'].get('id'):
        print('Loaded Project ID From: %s, Project Id' % CONNECTION_USER)
      else:
        # check service
        try:

          service_connection_extra = BaseHook.get_connection(CONNECTION_SERVICE).extra_dejson
          self.recipe['setup']['id'] = service_connection_extra.get('extra__google_cloud_platform__project')

          # check service json
          if self.recipe['setup'].get('id'):
            print('Loaded Project ID From: %s, Project Id' % CONNECTION_SERVICE)
          else:
            self.recipe['setup']['id'] = json.loads(
              service_connection_extra['extra__google_cloud_platform__keyfile_dict']
            )['project_id']
            print('Loaded Project ID From: %s, Keyfile JSON' % CONNECTION_SERVICE)

        except: pass


  def python_task(self, function, instance):
    print('STARTHINKER TASK:', function, instance)

    PythonOperator(
        task_id='%s_%d' % (function, instance),
        python_callable=getattr(
            import_module('starthinker.task.%s.run' % function), function),
        op_kwargs={
            'recipe': self.recipe,
            'instance': instance
        },
        dag=self.dag)

    return self


  def airflow_task(self, task, instance, parameters):
    print('STARTHINKER AIRFLOW TASK:', task, instance)

    def airflow_import_unroll(definition):
      return next(iter([i for i in definition.items() if i[0] != '__comment__']))

    af_path, af_module = airflow_import_unroll(parameters)
    af_module, af_operator = airflow_import_unroll(af_module)
    af_operator, af_parameters = airflow_import_unroll(af_operator)

    operator = getattr(import_module('%s.%s.%s' % (task, af_path, af_module)), af_operator)

    af_parameters['dag'] = self.dag
    af_parameters['task_id'] = '%s_%d' % (task, instance)

    operator(**af_parameters)

    return self


  def airflow_start(self):
    return datetime.now(self.recipe.get('setup', {}).get('timezone')) - timedelta(days=7)


  def airflow_schedule(self):
    recipe_day = self.recipe.get('setup', {}).get('day', [])
    recipe_hour = [str(h) for h in self.recipe.get('setup', {}).get('hour', [])]

    airflow_schedule = None
    if recipe_day or recipe_hour:
      airflow_schedule = '{minutes} {hours} {day_of_month} {month} {day_of_week}'.format(**{
        'minutes':randint(0, 15),
        'hours':','.join(recipe_hour) or '*',
        'day_of_month':'*',
        'month':'*',
        'day_of_week':','.join(recipe_day) or '*'
      })

    return airflow_schedule

  def generate(self):
    print('STARTHINKER DAG:', self.dag_name)

    self.dag = DAG(
        dag_id=self.dag_name,
        default_args={
            'owner': 'airflow',
            'start_date': self.airflow_start()
        },
        schedule_interval=self.airflow_schedule(),
        catchup=False)

    instances = {}
    for task in self.recipe['tasks']:
      function = next(iter(task.keys()))

      # count instance per task
      instances.setdefault(function, 0)
      instances[function] += 1

      # if airflow function ( product )
      if function == 'airflow':
        self.airflow_task(function, instances[function], task[function])

      # if airflow function ( local )
      elif function == 'starthinker.airflow':
        self.airflow_task(function, instances[function], task[function])

      # if native python operator
      else:
        self.python_task(function, instances[function])

    return self.dag


  def print_commandline(self):

    print('')
    print('STARTHINKER COMMANDLINE: %s' % self.dag_name)
    print('')

    instances = {}
    for task in self.recipe['tasks']:
      function = next(iter(task.keys()))

      # count instance per task
      instances.setdefault(function, 0)
      instances[function] += 1

      print('airflow test "%s" %s_%d %s' %
            (self.dag_name, function, instances[function], str(date.today())))
    print('')



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from starthinker.task.entity.schema.ExchangeSettings import ExchangeSettings_Schema
from starthinker.task.entity.schema.UniversalChannel import UniversalChannel_Schema
from starthinker.task.entity.schema.CostTrackingPixel import CostTrackingPixel_Schema
from starthinker.task.entity.schema.PartnerRevenueModel import PartnerRevenueModel_Schema
from starthinker.task.entity.schema.CustomAffinity import CustomAffinity_Schema
from starthinker.task.entity.schema.UniversalSite import UniversalSite_Schema
from starthinker.task.entity.schema.FrequencyCap import FrequencyCap_Schema
from starthinker.task.entity.schema.Creative import Creative_Schema
from starthinker.task.entity.schema.LineItem import LineItem_Schema
from starthinker.task.entity.schema.Pixel import Pixel_Schema
from starthinker.task.entity.schema.EntityCommonData import EntityCommonData_Schema
from starthinker.task.entity.schema.UserList import UserList_Schema
from starthinker.task.entity.schema.InventorySource import InventorySource_Schema
from starthinker.task.entity.schema.FreeFormTarget import FreeFormTarget_Schema
from starthinker.task.entity.schema.DeviceCriteria import DeviceCriteria_Schema
from starthinker.task.entity.schema.SupportedExchange import SupportedExchange_Schema
from starthinker.task.entity.schema.DataPartner import DataPartner_Schema
from starthinker.task.entity.schema.UserListPricing import UserListPricing_Schema
from starthinker.task.entity.schema.TargetUnion import TargetUnion_Schema
from starthinker.task.entity.schema.Target import Target_Schema
from starthinker.task.entity.schema.Language import Language_Schema
from starthinker.task.entity.schema.ApprovalStatus import ApprovalStatus_Schema
from starthinker.task.entity.schema.InsertionOrder import InsertionOrder_Schema
from starthinker.task.entity.schema.Budget import Budget_Schema
from starthinker.task.entity.schema.UserListAdvertiserPricing import UserListAdvertiserPricing_Schema
from starthinker.task.entity.schema.Browser import Browser_Schema
from starthinker.task.entity.schema.Advertiser import Advertiser_Schema
from starthinker.task.entity.schema.PartnerCosts import PartnerCosts_Schema
from starthinker.task.entity.schema.Campaign import Campaign_Schema
from starthinker.task.entity.schema.GeoLocation import GeoLocation_Schema
from starthinker.task.entity.schema.SelectionTarget import SelectionTarget_Schema
from starthinker.task.entity.schema.Isp import Isp_Schema
from starthinker.task.entity.schema.Partner import Partner_Schema
from starthinker.task.entity.schema.TargetList import TargetList_Schema

Entity_Schema_Lookup = {
    'ExchangeSettings': ExchangeSettings_Schema,
    'UniversalChannel': UniversalChannel_Schema,
    'CostTrackingPixel': CostTrackingPixel_Schema,
    'PartnerRevenueModel': PartnerRevenueModel_Schema,
    'CustomAffinity': CustomAffinity_Schema,
    'UniversalSite': UniversalSite_Schema,
    'FrequencyCap': FrequencyCap_Schema,
    'Creative': Creative_Schema,
    'LineItem': LineItem_Schema,
    'Pixel': Pixel_Schema,
    'EntityCommonData': EntityCommonData_Schema,
    'UserList': UserList_Schema,
    'InventorySource': InventorySource_Schema,
    'FreeFormTarget': FreeFormTarget_Schema,
    'DeviceCriteria': DeviceCriteria_Schema,
    'SupportedExchange': SupportedExchange_Schema,
    'DataPartner': DataPartner_Schema,
    'UserListPricing': UserListPricing_Schema,
    'TargetUnion': TargetUnion_Schema,
    'Target': Target_Schema,
    'Language': Language_Schema,
    'ApprovalStatus': ApprovalStatus_Schema,
    'InsertionOrder': InsertionOrder_Schema,
    'Budget': Budget_Schema,
    'UserListAdvertiserPricing': UserListAdvertiserPricing_Schema,
    'Browser': Browser_Schema,
    'Advertiser': Advertiser_Schema,
    'PartnerCosts': PartnerCosts_Schema,
    'Campaign': Campaign_Schema,
    'GeoLocation': GeoLocation_Schema,
    'SelectionTarget': SelectionTarget_Schema,
    'Isp': Isp_Schema,
    'Partner': Partner_Schema,
    'TargetList': TargetList_Schema,
}



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################
"""Handles creation and updates of Placements."""

from starthinker.task.traffic.dao import BaseDAO
from starthinker.task.traffic.campaign import CampaignDAO
from starthinker.task.traffic.video_format import VideoFormatDAO
from starthinker.task.traffic.feed import FieldMap
from starthinker.task.traffic.placement_group import PlacementGroupDAO
from starthinker.task.traffic.class_extensions import StringExtensions


class PlacementDAO(BaseDAO):
  """Placement data access object.

  Inherits from BaseDAO and implements placement specific logic for creating and
  updating placement.
  """

  cache = {}

  def __init__(self, auth, profile_id, is_admin):
    """Initializes PlacementDAO with profile id and authentication scheme."""
    super(PlacementDAO, self).__init__(auth, profile_id, is_admin)

    self._entity = 'PLACEMENT'

    self.campaign_dao = CampaignDAO(auth, profile_id, is_admin)
    self.video_format_dao = VideoFormatDAO(auth, profile_id, is_admin)
    self.placement_group_dao = PlacementGroupDAO(auth, profile_id, is_admin)

    self._id_field = FieldMap.PLACEMENT_ID
    self._search_field = FieldMap.PLACEMENT_NAME

    self._parent_filter_name = 'campaignIds'
    self._parent_filter_field_name = FieldMap.CAMPAIGN_ID
    self._parent_dao = self.campaign_dao

    self._list_name = 'placements'

    self.cache = PlacementDAO.cache

  def _api(self, iterate=False):
    """Returns an DCM API instance for this DAO."""
    return super(PlacementDAO, self)._api(iterate).placements()

  def _api_sizes(self, iterate=False):
    """Returns an DCM API instance for this DAO."""
    return super(PlacementDAO, self)._api(iterate).sizes()

  def _process_skipability(self, feed_item, item):
    """Process skipability settings.

    Args:
      feed_item: A feed item representing a placement from the bulkdozer feed;
      item: A campaign manager placement object to be updated with the
        skipability settings defined in the feed item
    """
    if feed_item.get(FieldMap.PLACEMENT_SKIPPABLE, False):
      if not 'videoSettings' in item:
        item['videoSettings'] = {}

      item['videoSettings']['skippableSettings'] = {
          'skippable': feed_item.get(FieldMap.PLACEMENT_SKIPPABLE, False),
          'skipOffset': {},
          'progressOffset': {}
      }

      skippable_settings = item['videoSettings']['skippableSettings']

      if feed_item.get(FieldMap.PLACEMENT_SKIP_OFFSET_SECONDS, None):
        skippable_settings['skipOffset']['offsetSeconds'] = feed_item.get(
            FieldMap.PLACEMENT_SKIP_OFFSET_SECONDS, None)

      if feed_item.get(FieldMap.PLACEMENT_SKIP_OFFSET_PERCENTAGE, None):
        skippable_settings['skipOffset']['offsetPercentage'] = feed_item.get(
            FieldMap.PLACEMENT_SKIP_OFFSET_PERCENTAGE, None)

      if feed_item.get(FieldMap.PLACEMENT_PROGRESS_OFFSET_SECONDS, None):
        skippable_settings['progressOffset']['offsetSeconds'] = feed_item.get(
            FieldMap.PLACEMENT_SKIP_OFFSET_SECONDS, None)

      if feed_item.get(FieldMap.PLACEMENT_PROGRESS_OFFSET_PERCENTAGE, None):
        skippable_settings['progressOffset'][
            'offsetPercentage'] = feed_item.get(
                FieldMap.PLACEMENT_SKIP_OFFSET_PERCENTAGE, None)
    else:
      if 'skippableSettings' in item and 'videoSettings' in item:
        del item['videoSettings']['skippableSettings']

  def _process_active_view_and_verification(self, placement, feed_item):
    """Updates / creates active view and verification settings.

    This method updates the CM item by setting or creating active view and
    verification settings based on the Bulkdozer feed configurations.

    Args:
      placement: The CM placement object to be updated.
      feed_item: The Bulkdozer feed item with the configurations.

    Raises:
      Exception: In case the values for active view and verification enumeration
      is invalid.
    """

    if FieldMap.PLACEMENT_ACTIVE_VIEW_AND_VERIFICATION in feed_item:
      if feed_item.get(FieldMap.PLACEMENT_ACTIVE_VIEW_AND_VERIFICATION,
                       None) == 'ON':
        placement['vpaidAdapterChoice'] = 'HTML5'
        placement['videoActiveViewOptOut'] = False
      elif feed_item.get(FieldMap.PLACEMENT_ACTIVE_VIEW_AND_VERIFICATION,
                         None) == 'OFF':
        placement['vpaidAdapterChoice'] = 'DEFAULT'
        placement['videoActiveViewOptOut'] = True
      elif feed_item[
          FieldMap.
          PLACEMENT_ACTIVE_VIEW_AND_VERIFICATION] == 'LET_DCM_DECIDE' or feed_item[
              FieldMap.PLACEMENT_ACTIVE_VIEW_AND_VERIFICATION] == '':
        placement['vpaidAdapterChoice'] = 'DEFAULT'
        placement['videoActiveViewOptOut'] = False
      else:
        raise Exception(
            '%s is not a valid value for the placement Active View and Verification field'
            % feed_item.get(FieldMap.PLACEMENT_ACTIVE_VIEW_AND_VERIFICATION,
                            None))

  def _process_pricing_schedule(self, item, feed_item):
    """Updates / creates pricing schedule settings.

    This method updates the CM item with pricing schedule based on
    configurations from the Bulkdozer feed.

    Args:
      item: the CM placement object to update.
      feed_item: The Bulkdozer feed item representing the settings to define.
    """
    if 'pricing_schedule' in feed_item and feed_item['pricing_schedule']:
      if not 'pricingSchedule' in item:
        item['pricingSchedule'] = {}

      item['pricingSchedule']['pricingPeriods'] = []

      for pricing_schedule in feed_item['pricing_schedule']:
        item['pricingSchedule']['pricingPeriods'].append({
            'endDate':
                pricing_schedule.get(FieldMap.PLACEMENT_PERIOD_END, None),
            'startDate':
                pricing_schedule.get(FieldMap.PLACEMENT_PERIOD_START, None),
            'rateOrCostNanos':
                int(
                    float(pricing_schedule.get(FieldMap.PLACEMENT_PERIOD_RATE))
                    * 1000000000),
            'units':
                pricing_schedule.get(FieldMap.PLACEMENT_PERIOD_UNITS),
        })

  def _process_update(self, item, feed_item):
    """Updates an placement based on the values from the feed.

    Args:
      item: Object representing the placement to be updated, this object is
        updated directly.
      feed_item: Feed item representing placement values from the Bulkdozer
        feed.
    """

    if feed_item.get(FieldMap.CAMPAIGN_ID, '') == '':
      feed_item[FieldMap.CAMPAIGN_ID] = item['campaignId']

    campaign = self.campaign_dao.get(feed_item, required=True)
    placement_group = self.placement_group_dao.get(feed_item, required=True)

    feed_item[FieldMap.CAMPAIGN_ID] = campaign['id']
    feed_item[FieldMap.CAMPAIGN_NAME] = campaign['name']

    if placement_group:
      feed_item[FieldMap.PLACEMENT_GROUP_ID] = placement_group['id']
      feed_item[FieldMap.PLACEMENT_GROUP_NAME] = placement_group['name']
      item['placementGroupId'] = placement_group['id']
    else:
      item['placementGroupId'] = None

    self._process_skipability(feed_item, item)

    item['pricingSchedule']['startDate'] = (
        StringExtensions.convertDateTimeStrToDateStr(
            feed_item.get(FieldMap.PLACEMENT_START_DATE, None))
        if feed_item.get(FieldMap.PLACEMENT_START_DATE, '') else
        item['pricingSchedule']['startDate'])

    item['pricingSchedule']['endDate'] = (
        StringExtensions.convertDateTimeStrToDateStr(
            feed_item.get(FieldMap.PLACEMENT_END_DATE, None)) if feed_item.get(
                FieldMap.PLACEMENT_END_DATE, '') else
        item['pricingSchedule']['endDate'])

    item['pricingSchedule']['pricingType'] = feed_item.get(
        FieldMap.PLACEMENT_PRICING_SCHEDULE_COST_STRUCTURE,
        None) if feed_item.get(
            FieldMap.PLACEMENT_PRICING_SCHEDULE_COST_STRUCTURE,
            '') else item['pricingSchedule']['pricingType']

    if feed_item.get(FieldMap.PLACEMENT_PRICING_TESTING_START, None):
      item['pricingSchedule']['testingStartDate'] = feed_item.get(
          FieldMap.PLACEMENT_PRICING_TESTING_START, None)

    item['name'] = feed_item.get(FieldMap.PLACEMENT_NAME,
                                 None) if feed_item.get(FieldMap.PLACEMENT_NAME,
                                                        '') else item['name']
    item['archived'] = feed_item.get(
        FieldMap.PLACEMENT_ARCHIVED, None) if feed_item.get(
            FieldMap.PLACEMENT_ARCHIVED, '') else item['archived']
    item['adBlockingOptOut'] = feed_item.get(FieldMap.PLACEMENT_AD_BLOCKING,
                                             False)

    self._process_transcode(item, feed_item)
    self._process_active_view_and_verification(item, feed_item)
    self._process_pricing_schedule(item, feed_item)

    if feed_item.get(FieldMap.PLACEMENT_ADDITIONAL_KEY_VALUES, None):
      if not 'tagSetting' in item:
        item['tagSetting'] = {}

      item['tagSetting']['additionalKeyValues'] = feed_item.get(
          FieldMap.PLACEMENT_ADDITIONAL_KEY_VALUES, None)

  def _process_transcode(self, item, feed_item):
    """Updates / creates transcode configuration for the placement.

    This method updates the CM placement object with transcoding configuration
    from the feed.

    Args:
      item: The CM placement object to update.
      feed_item: The Bulkdozer feed item with the transcode configurations.
    """
    if feed_item.get('transcode_config', None):
      if not 'videoSettings' in item:
        item['videoSettings'] = {}

      if not 'transcodeSettings' in item['videoSettings']:
        item['videoSettings']['transcodeSettings'] = {}

      item['videoSettings']['transcodeSettings'][
          'enabledVideoFormats'] = self.video_format_dao.translate_transcode_config(
              feed_item['transcode_config'])

      if not item['videoSettings']['transcodeSettings']['enabledVideoFormats']:
        raise Exception(
            'Specified transcode profile did not match any placement level transcode settings in Campaign Manager'
        )

  def get_sizes(self, width, height):
    """Retrieves a creative sizes from DCM.

    Args:
      width: width of the creative.
      height: height of the creative.
      retry_count: how many times the api call should be retried in case of an
        api related error that is not a client error.

    Returns:
      The sizes object from DCM.
    """
    # TODO (mauriciod): this could potentially be in a separate SizesDAO,
    # but since we don't use it anywhere else it is probably fine.
    # May need to do it in case it becomes necessary for other entities when
    # we implement display
    return list(
        self._api_sizes(iterate=True).list(
            profileId=self.profile_id, height=height, width=width).execute())

  def _process_new(self, feed_item):
    """Creates a new placement DCM object from a feed item representing an placement from the Bulkdozer feed.

    This function simply creates the object to be inserted later by the BaseDAO
    object.

    Args:
      feed_item: Feed item representing the placement from the Bulkdozer feed.

    Returns:
      An placement object ready to be inserted in DCM through the API.

    """
    campaign = self.campaign_dao.get(feed_item, required=True)
    placement_group = self.placement_group_dao.get(feed_item, required=True)

    feed_item[FieldMap.CAMPAIGN_ID] = campaign['id']
    feed_item[FieldMap.CAMPAIGN_NAME] = campaign['name']

    if placement_group:
      feed_item[FieldMap.PLACEMENT_GROUP_ID] = placement_group['id']
      feed_item[FieldMap.PLACEMENT_GROUP_NAME] = placement_group['name']

    result = {
        'name':
            feed_item.get(FieldMap.PLACEMENT_NAME, None),
        'adBlockingOptOut':
            feed_item.get(FieldMap.PLACEMENT_AD_BLOCKING, False),
        'campaignId':
            campaign['id'],
        'placementGroupId':
            placement_group['id'] if placement_group else None,
        'archived':
            feed_item.get(FieldMap.PLACEMENT_ARCHIVED, False),
        'siteId':
            feed_item.get(FieldMap.SITE_ID, None),
        'paymentSource':
            'PLACEMENT_AGENCY_PAID',
        'pricingSchedule': {
            'startDate':
                StringExtensions.convertDateTimeStrToDateStr(
                    feed_item.get(FieldMap.PLACEMENT_START_DATE, None)),
            'endDate':
                StringExtensions.convertDateTimeStrToDateStr(
                    feed_item.get(FieldMap.PLACEMENT_END_DATE, None)),
            'pricingType':
                feed_item.get(
                    FieldMap.PLACEMENT_PRICING_SCHEDULE_COST_STRUCTURE, None)
                or 'PRICING_TYPE_CPM',
            'pricingPeriods': [{
                'startDate': feed_item.get(FieldMap.PLACEMENT_START_DATE, None),
                'endDate': feed_item.get(FieldMap.PLACEMENT_END_DATE, None)
            }]
        }
    }

    self._process_skipability(feed_item, result)

    if feed_item.get(FieldMap.PLACEMENT_ADDITIONAL_KEY_VALUES, None):
      result['tagSetting'] = {
          'additionalKeyValues':
              feed_item.get(FieldMap.PLACEMENT_ADDITIONAL_KEY_VALUES, None)
      }

    if feed_item.get(FieldMap.PLACEMENT_PRICING_TESTING_START, None):
      result['pricingSchedule']['testingStartDate'] = feed_item.get(
          FieldMap.PLACEMENT_PRICING_TESTING_START, None)

    self._process_active_view_and_verification(result, feed_item)

    if feed_item.get(FieldMap.PLACEMENT_TYPE, None) == 'VIDEO' or feed_item[
        FieldMap.PLACEMENT_TYPE] == 'IN_STREAM_VIDEO':
      result['compatibility'] = 'IN_STREAM_VIDEO'
      result['size'] = {'width': '0', 'height': '0'}
      result['tagFormats'] = ['PLACEMENT_TAG_INSTREAM_VIDEO_PREFETCH']
    elif feed_item[FieldMap.PLACEMENT_TYPE] == 'IN_STREAM_AUDIO':
       result['compatibility'] = 'IN_STREAM_AUDIO'
       result['size'] = {'width': '0', 'height': '0'}
       result['tagFormats'] = ['PLACEMENT_TAG_INSTREAM_VIDEO_PREFETCH']
    else:
      result['compatibility'] = 'DISPLAY'
      width = 1
      height = 1
      raw_size = feed_item.get(FieldMap.ASSET_SIZE, '0x0')
      if (raw_size and 'x' in raw_size):
        width, height = raw_size.strip().lower().split('x')

      sizes = self.get_sizes(int(width), int(height))

      if sizes:
        result['size'] = {'id': sizes[0]['id']}
      else:
        result['size'] = {'width': int(width), 'height': int(height)}

      result['tagFormats'] = [
          'PLACEMENT_TAG_STANDARD', 'PLACEMENT_TAG_JAVASCRIPT',
          'PLACEMENT_TAG_IFRAME_JAVASCRIPT', 'PLACEMENT_TAG_IFRAME_ILAYER',
          'PLACEMENT_TAG_INTERNAL_REDIRECT', 'PLACEMENT_TAG_TRACKING',
          'PLACEMENT_TAG_TRACKING_IFRAME', 'PLACEMENT_TAG_TRACKING_JAVASCRIPT'
      ]

    self._process_transcode(result, feed_item)
    self._process_pricing_schedule(result, feed_item)

    return result

  def _post_process(self, feed_item, item):

    for pricing_schedule in feed_item.get('pricing_schedule', []):
      placement = self.get(pricing_schedule)

      if placement:
        feed_item[FieldMap.PLACEMENT_ID] = placement['id']

  def map_placement_transcode_configs(self, placement_feed,
                                      transcode_configs_feed,
                                      pricing_schedule_feed):
    """Maps sub feeds with the parent feed based on placement id.

    Args:
      placement_feed: Bulkdozer feed representing the placements configurations.
      trascode_configs_feed: Bulkdozer feed representing the transcode configs.
      pricing_schedule_feed: Bulkdozer feed representing the pricing schedules.
    """

    for placement in placement_feed:
      placement['pricing_schedule'] = []

      for pricing_schedule in pricing_schedule_feed:
        if placement.get(FieldMap.PLACEMENT_ID,
                         '') == pricing_schedule.get(FieldMap.PLACEMENT_ID,
                                                     None):
          placement['pricing_schedule'].append(pricing_schedule)

      transcode_id = placement.get(FieldMap.TRANSCODE_ID, '')
      placement['transcode_config'] = []
      if transcode_id:
        for transcode_config in transcode_configs_feed:
          if transcode_id == transcode_config.get(FieldMap.TRANSCODE_ID, None):
            placement['transcode_config'].append(transcode_config)



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################
import re

from starthinker.util.project import project
from starthinker.util.sheets import sheets_read, sheets_tab_copy
from starthinker.util.bigquery import get_schema, query_to_view
from starthinker.util.csv import rows_to_type, rows_to_csv

TEMPLATE_SHEET = 'https://docs.google.com/spreadsheets/d/1_faknNlaPlltLwsleMyQH0Gis5ixJODLGJHLovY3Ycg/edit?usp=sharing'
TEMPLATE_TAB = 'Mapping'
RE_SQLINJECT = re.compile(r'[^a-z0-9_\-, ]+', re.UNICODE | re.IGNORECASE)


@project.from_parameters
def mapping():
  if project.verbose:
    print('MAPPING')

  # create the sheet from template if it does not exist
  sheets_tab_copy(project.task['auth'], TEMPLATE_SHEET, TEMPLATE_TAB,
                  project.task['sheet'], project.task['tab'])

  # move if specified
  dimensions = {}
  defaults = {}
  rows = sheets_read(project.task['auth'], project.task['sheet'],
                     project.task['tab'], 'A1:D')

  # if rows don't exist, query is still created without mapping ( allows blank maps )
  if rows:
    # sanitize mapping
    # 0 = Dimension, 1 = Tag, 2 = Column, 3 = Keyword
    for row in rows[1:]:
      if project.verbose:
        print('ROW: ', row)
      # sanitize row
      #row = map(lambda c: RE_SQLINJECT.sub('', c.strip()), row)
      row = [RE_SQLINJECT.sub('', r.strip()) for r in row]
      if len(row) == 2:  # default
        defaults.setdefault(row[0], row[1])
      else:  # tag
        dimensions.setdefault(row[0], {})  # dimension
        dimensions[row[0]].setdefault(row[1], {})
        dimensions[row[0]].setdefault(row[1], {})  # tag
        dimensions[row[0]][row[1]].setdefault(row[2], [])  # column
        dimensions[row[0]][row[1]][row[2]].extend(
            [k.strip() for k in row[3].split(',') if k])  # keywords

  # construct query
  query = 'SELECT\n  *,\n'
  for dimension, tags in dimensions.items():
    query += '  CASE\n'
    for tag, columns in tags.items():
      query += '    WHEN '
      for column, keywords in columns.items():
        for count, keyword in enumerate(keywords):
          if count != 0:
            query += 'OR '
          query += '%s CONTAINS "%s" ' % (column, keyword)
      query += 'THEN "%s"\n' % tag
    query += '    ELSE "%s"\n  END AS %s,\n' % (defaults.get(dimension,
                                                             ''), dimension)
  query += 'FROM [%s.%s]' % (project.task['in']['dataset'],
                             project.task['in']['table'])

  if project.verbose:
    print('QUERY: ', query)

  # write to view
  query_to_view(
      project.task['out']['auth'],
      project.id,
      project.task['out']['dataset'],
      project.task['out']['view'],
      query,
      replace=True)


if __name__ == '__main__':
  mapping()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from starthinker.util.bigquery import query_to_view
from starthinker.util.bigquery import table_create
from starthinker.util.data import get_rows
from starthinker.util.data import put_rows
from starthinker.util.google_api.discovery_to_bigquery import Discovery_To_BigQuery
from starthinker.util.project import project
from starthinker.util.regexp import lookup_id
from starthinker.util.sheets import sheets_clear

from starthinker.task.dv_editor.insertion_order import insertion_order_commit
from starthinker.task.dv_editor.line_item import line_item_commit
from starthinker.task.dv_editor.patch import patch_masks
from starthinker.task.dv_editor.patch import patch_preview


def bid_strategy_clear():
  sheets_clear(project.task["auth_sheets"], project.task["sheet"], "Bid Strategy",
               "A2:Z")


def bid_strategy_load():

  # write bid_strategy to sheet
  rows = get_rows(
      project.task["auth_bigquery"], {
          "bigquery": {
              "dataset":
                  project.task["dataset"],
              "query":
                  """SELECT
         CONCAT(P.displayName, ' - ', P.partnerId),
         CONCAT(A.displayName, ' - ', A.advertiserId),
         CONCAT(C.displayName, ' - ', C.campaignId),
         CONCAT(I.displayName, ' - ', I.insertionOrderId),
         NULL,
         I.bidStrategy.fixedBid.bidAmountMicros / 100000,
         I.bidStrategy.fixedBid.bidAmountMicros / 100000,
         I.bidStrategy.maximizeSpendAutoBid.performanceGoalType,
         I.bidStrategy.maximizeSpendAutoBid.performanceGoalType,
         I.bidStrategy.maximizeSpendAutoBid.maxAverageCpmBidAmountMicros / 100000,
         I.bidStrategy.maximizeSpendAutoBid.maxAverageCpmBidAmountMicros / 100000,
         I.bidStrategy.maximizeSpendAutoBid.customBiddingAlgorithmId,
         I.bidStrategy.maximizeSpendAutoBid.customBiddingAlgorithmId,
         I.bidStrategy.performanceGoalAutoBid.performanceGoalType,
         I.bidStrategy.performanceGoalAutoBid.performanceGoalType,
         I.bidStrategy.performanceGoalAutoBid.performanceGoalAmountMicros / 100000,
         I.bidStrategy.performanceGoalAutoBid.performanceGoalAmountMicros / 100000,
         I.bidStrategy.performanceGoalAutoBid.maxAverageCpmBidAmountMicros / 100000,
         I.bidStrategy.performanceGoalAutoBid.maxAverageCpmBidAmountMicros / 100000,
         I.bidStrategy.performanceGoalAutoBid.customBiddingAlgorithmId,
         I.bidStrategy.performanceGoalAutoBid.customBiddingAlgorithmId
       FROM `{dataset}.DV_InsertionOrders` AS I
       LEFT JOIN `{dataset}.DV_Campaigns` AS C
       ON I.campaignId=C.campaignId
       LEFT JOIN `{dataset}.DV_Advertisers` AS A
       ON I.advertiserId=A.advertiserId
       LEFT JOIN `{dataset}.DV_Partners` AS P
       ON A.partnerId=P.partnerId
       UNION ALL
       SELECT
         CONCAT(P.displayName, ' - ', P.partnerId),
         CONCAT(A.displayName, ' - ', A.advertiserId),
         CONCAT(C.displayName, ' - ', C.campaignId),
         CONCAT(I.displayName, ' - ', I.insertionOrderId),
         CONCAT(L.displayName, ' - ', L.lineItemId),
         L.bidStrategy.fixedBid.bidAmountMicros / 100000,
         L.bidStrategy.fixedBid.bidAmountMicros / 100000,
         L.bidStrategy.maximizeSpendAutoBid.performanceGoalType,
         L.bidStrategy.maximizeSpendAutoBid.performanceGoalType,
         L.bidStrategy.maximizeSpendAutoBid.maxAverageCpmBidAmountMicros / 100000,
         L.bidStrategy.maximizeSpendAutoBid.maxAverageCpmBidAmountMicros / 100000,
         L.bidStrategy.maximizeSpendAutoBid.customBiddingAlgorithmId,
         L.bidStrategy.maximizeSpendAutoBid.customBiddingAlgorithmId,
         L.bidStrategy.performanceGoalAutoBid.performanceGoalType,
         L.bidStrategy.performanceGoalAutoBid.performanceGoalType,
         L.bidStrategy.performanceGoalAutoBid.performanceGoalAmountMicros / 100000,
         L.bidStrategy.performanceGoalAutoBid.performanceGoalAmountMicros / 100000,
         L.bidStrategy.performanceGoalAutoBid.maxAverageCpmBidAmountMicros / 100000,
         L.bidStrategy.performanceGoalAutoBid.maxAverageCpmBidAmountMicros / 100000,
         L.bidStrategy.performanceGoalAutoBid.customBiddingAlgorithmId,
         L.bidStrategy.performanceGoalAutoBid.customBiddingAlgorithmId
       FROM `{dataset}.DV_LineItems` AS L
       LEFT JOIN `{dataset}.DV_Campaigns` AS C
       ON L.campaignId=C.campaignId
       LEFT JOIN `{dataset}.DV_InsertionOrders` AS I
       ON L.insertionOrderId=I.insertionOrderId
       LEFT JOIN `{dataset}.DV_Advertisers` AS A
       ON L.advertiserId=A.advertiserId
       LEFT JOIN `{dataset}.DV_Partners` AS P
       ON A.partnerId=P.partnerId
       """.format(**project.task),
              "legacy":
                  False
          }
      })

  put_rows(
      project.task["auth_sheets"], {
          "sheets": {
              "sheet": project.task["sheet"],
              "tab": "Bid Strategy",
              "range": "A2"
          }
      }, rows)


def bid_strategy_audit():
  rows = get_rows(
      project.task["auth_sheets"], {
          "sheets": {
              "sheet": project.task["sheet"],
              "tab": "Bid Strategy",
              "range": "A2:Z"
          }
      })

  put_rows(
      project.task["auth_bigquery"], {
          "bigquery": {
              "dataset": project.task["dataset"],
              "table": "SHEET_BidStrategy",
              "schema": [
                  { "name": "Partner", "type": "STRING" },
                  { "name": "Advertiser", "type": "STRING" },
                  { "name": "Campaign", "type": "STRING" },
                  { "name": "Insertion_Order", "type": "STRING" },
                  { "name": "Line_Item", "type": "STRING" },
                  { "name": "Fixed_Bid", "type": "FLOAT" },
                  { "name": "Fixed_Bid_Edit", "type": "FLOAT" },
                  { "name": "Auto_Bid_Goal", "type": "STRING" },
                  { "name": "Auto_Bid_Goal_Edit", "type": "STRING" },
                  { "name": "Auto_Bid_Amount", "type": "FLOAT" },
                  { "name": "Auto_Bid_Amount_Edit", "type": "FLOAT" },
                  { "name": "Auto_Bid_Algorithm", "type": "STRING" },
                  { "name": "Auto_Bid_Algorithm_Edit", "type": "STRING" },
                  { "name": "Performance_Goal_Type", "type": "STRING" },
                  { "name": "Performance_Goal_Type_Edit", "type": "STRING" },
                  { "name": "Performance_Goal_Amount", "type": "FLOAT" },
                  { "name": "Performance_Goal_Amount_Edit", "type": "FLOAT" },
                  { "name": "Performance_Goal_Average_CPM_Bid", "type": "FLOAT" },
                  { "name": "Performance_Goal_Average_CPM_Bid_Edit", "type": "FLOAT" },
                  { "name": "Performance_Goal_Algorithm", "type": "STRING" },
                  { "name": "Performance_Goal_Algorithm_Edit", "type": "STRING" },
              ],
              "format": "CSV"
          }
      }, rows)

  query_to_view(
      project.task["auth_bigquery"],
      project.id,
      project.task["dataset"],
      "AUDIT_BidStrategy",
      """WITH
      /* Check if sheet values are set */
      INPUT_ERRORS AS (
        SELECT
        *
        FROM (
          SELECT
            'Bid Strategy' AS Operation,
            CASE
              WHEN Fixed_Bid_Edit IS NULL THEN 'Missing Fixed Bid.'
              WHEN Auto_Bid_Goal_Edit IS NULL THEN 'Missing Auto Bid Goal.'
              WHEN Auto_Bid_Algorithm_Edit IS NULL THEN 'Missing Auto Bid Algorithm.'
            ELSE
              NULL
            END AS Error,
            'ERROR' AS Severity,
          COALESCE(Line_Item, Insertion_Order, 'BLANK') AS Id
        FROM
          `{dataset}.SHEET_BidStrategy`
        )
        WHERE
          Error IS NOT NULL
      )

      SELECT * FROM INPUT_ERRORS
      ;
    """.format(**project.task),
      legacy=False)

  query_to_view(
    project.task["auth_bigquery"],
    project.id,
    project.task["dataset"],
    "PATCH_BidStrategy",
    """SELECT *
      FROM `{dataset}.SHEET_BidStrategy`
      WHERE (
        REGEXP_CONTAINS(Insertion_Order, r" - (\d+)$")
        OR REGEXP_CONTAINS(Line_Item, r" - (\d+)$")
      )
      AND Line_Item NOT IN (SELECT Id FROM `{dataset}.AUDIT_BidStrategy` WHERE Severity='ERROR')
      AND Insertion_Order NOT IN (SELECT Id FROM `{dataset}.AUDIT_BidStrategy` WHERE Severity='ERROR')
    """.format(**project.task),
    legacy=False
  )


def bid_strategy_patch(commit=False):
  patches = []

  rows = get_rows(
    project.task["auth_bigquery"],
    { "bigquery": {
      "dataset": project.task["dataset"],
      "table":"PATCH_BidStrategy",
    }},
    as_object=True
  )

  for row in rows:

    bid_strategy = {}

    if row['Fixed_Bid'] != row['Fixed_Bid_Edit']:
      bid_strategy.setdefault("bidStrategy", {"fixedBid": {}})
      bid_strategy["bidStrategy"]["fixedBid"]["bidAmountMicros"] = int(
        float(row['Fixed_Bid_Edit']) * 100000
      )

    if row['Auto_Bid_Goal'] != row['Auto_Bid_Goal_Edit']:
      bid_strategy.setdefault("bidStrategy", {"maximizeSpendAutoBid": {}})
      bid_strategy["bidStrategy"]["maximizeSpendAutoBid"][
       "performanceGoalType"] = row['Auto_Bid_Goal_Edit']

    if row['Auto_Bid_Amount'] != row['Auto_Bid_Amount_Edit']:
      bid_strategy.setdefault("bidStrategy", {"maximizeSpendAutoBid": {}})
      bid_strategy["bidStrategy"]["maximizeSpendAutoBid"][
        "maxAverageCpmBidAmountMicros"] = int(float(row['Auto_Bid_Amount_Edit']) * 100000)

    if row['Auto_Bid_Algorithm'] != row['Auto_Bid_Algorithm_Edit']:
      bid_strategy.setdefault("bidStrategy", {"maximizeSpendAutoBid": {}})
      bid_strategy["bidStrategy"]["maximizeSpendAutoBid"][
        "customBiddingAlgorithmId"] = row['Auto_Bid_Algorithm_Edit']

    if row['Performance_Goal_Type'] != row['Performance_Goal_Type_Edit']:
      bid_strategy.setdefault("bidStrategy", {"performanceGoalAutoBid": {}})
      bid_strategy["bidStrategy"]["performanceGoalAutoBid"][
          "performanceGoalType"] = row['Performance_Goal_Type_Edit']

    if row['Performance_Goal_Amount'] != row['Performance_Goal_Amount_Edit']:
      bid_strategy.setdefault("bidStrategy", {"performanceGoalAutoBid": {}})
      bid_strategy["bidStrategy"]["performanceGoalAutoBid"][
          "performanceGoalAmountMicros"] = int(float(row['Performance_Goal_Amount_Edit']) * 100000)

    if row['Performance_Goal_Average_CPM_Bid'] != row['Performance_Goal_Average_CPM_Bid_Edit']:
      bid_strategy.setdefault("bidStrategy", {"performanceGoalAutoBid": {}})
      bid_strategy["bidStrategy"]["performanceGoalAutoBid"][
          "maxAverageCpmBidAmountMicros"] = int(float(row['Performance_Goal_Average_CPM_Bid_Edit']) * 100000)

    if row['Performance_Goal_Algorithm'] != row['Performance_Goal_Algorithm_Edit']:
      bid_strategy.setdefault("bidStrategy", {"performanceGoalAutoBid": {}})
      bid_strategy["bidStrategy"]["performanceGoalAutoBid"][
          "customBiddingAlgorithmId"] = row['Performance_Goal_Algorithm_Edit']

    if bid_strategy:
      patch = {
          "operation": "Bid Strategy",
          "action": "PATCH",
          "partner": row['Partner'],
          "advertiser": row['Advertiser'],
          "campaign": row['Campaign'],
          "parameters": {
              "advertiserId": lookup_id(row['Advertiser']),
              "body": bid_strategy
          }
      }

      if row['Line_Item']:
        patch["line_item"] = row['Line_Item']
        patch["parameters"]["lineItemId"] = lookup_id(row['Line_Item'])
      else:
        patch["insertion_order"] = row['Insertion_Order']
        patch["parameters"]["insertionOrderId"] = lookup_id(row['Insertion_Order'])

      patches.append(patch)

  patch_masks(patches)

  if commit:
    insertion_order_commit(patches)
    line_item_commit(patches)
  else:
    patch_preview(patches)



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################
"""Handles creation and updates of Ads."""

from starthinker.task.traffic.dao import BaseDAO
from starthinker.task.traffic.landing_page import LandingPageDAO
from starthinker.task.traffic.feed import FieldMap
from starthinker.task.traffic.store import store
from starthinker.task.traffic.class_extensions import StringExtensions


class CampaignDAO(BaseDAO):
  """Campaign data access object.

  Inherits from BaseDAO and implements campaign specific logic for creating and
  updating campaigns.
  """

  def __init__(self, auth, profile_id, is_admin):
    """Initializes CampaignDAO with profile id and authentication scheme."""
    super(CampaignDAO, self).__init__(auth, profile_id, is_admin)

    self.landing_page_dao = LandingPageDAO(auth, profile_id, is_admin)
    self._id_field = FieldMap.CAMPAIGN_ID
    self._search_field = FieldMap.CAMPAIGN_NAME
    self._list_name = 'campaigns'

    self._parent_filter_name = None
    self._parent_filter_field_name = None

    self._entity = 'CAMPAIGN'

  def _api(self, iterate=False):
    """Returns an DCM API instance for this DAO."""
    return super(CampaignDAO, self)._api(iterate).campaigns()

  def _process_update(self, item, feed_item):
    """Updates a campaign based on the values from the feed.

    Args:
      item: Object representing the campaign to be updated, this object is
        updated directly.
      feed_item: Feed item representing campaign values from the Bulkdozer feed.
    """
    lp = self.landing_page_dao.get(feed_item, required=True)

    feed_item[FieldMap.CAMPAIGN_LANDING_PAGE_ID] = lp['id']
    feed_item[FieldMap.CAMPAIGN_LANDING_PAGE_NAME] = lp['name']

    item['startDate'] = StringExtensions.convertDateTimeStrToDateStr(
        feed_item.get(FieldMap.CAMPAIGN_START_DATE, None))
    item['endDate'] = StringExtensions.convertDateTimeStrToDateStr(
        feed_item.get(FieldMap.CAMPAIGN_END_DATE, None))
    item['name'] = feed_item.get(FieldMap.CAMPAIGN_NAME, None)
    item['defaultLandingPageId'] = lp['id']

  def _process_new(self, feed_item):
    """Creates a new campaign DCM object from a feed item representing a campaign from the Bulkdozer feed.

    This function simply creates the object to be inserted later by the BaseDAO
    object.

    Args:
      feed_item: Feed item representing the campaign from the Bulkdozer feed.

    Returns:
      A campaign object ready to be inserted in DCM through the API.

    """
    lp = self.landing_page_dao.get(feed_item, required=True)

    feed_item[FieldMap.CAMPAIGN_LANDING_PAGE_ID] = lp['id']
    feed_item[FieldMap.CAMPAIGN_LANDING_PAGE_NAME] = lp['name']

    return {
        'advertiserId':
            feed_item.get(FieldMap.ADVERTISER_ID, None),
        'name':
            feed_item.get(FieldMap.CAMPAIGN_NAME, None),
        'startDate':
            StringExtensions.convertDateTimeStrToDateStr(
                feed_item.get(FieldMap.CAMPAIGN_START_DATE, None)),
        'endDate':
            StringExtensions.convertDateTimeStrToDateStr(
                feed_item.get(FieldMap.CAMPAIGN_END_DATE, None)),
        'defaultLandingPageId':
            lp['id']
    }



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

CUSTOM_DIMENSION_SCHEMA = [{
    'name': 'accountName',
    'type': 'STRING'
}, {
    'name': 'accountId',
    'type': 'STRING'
}, {
    'name': 'propertyName',
    'type': 'STRING'
}, {
    'name': 'propertyId',
    'type': 'STRING'
}, {
    'name': 'id',
    'type': 'STRING'
}, {
    'name': 'name',
    'type': 'STRING'
}, {
    'name': 'index',
    'type': 'STRING'
}, {
    'name': 'scope',
    'type': 'STRING'
}, {
    'name': 'active',
    'type': 'STRING'
}, {
    'name': 'created',
    'type': 'TIMESTAMP'
}, {
    'name': 'updated',
    'type': 'TIMESTAMP'
}, {
    'name': 'date',
    'type': 'DATE'
}, {
    'name': 'selfLink',
    'type': 'STRING'
}]

CUSTOM_METRIC_SCHEMA = [{
    'name': 'accountName',
    'type': 'STRING'
}, {
    'name': 'accountId',
    'type': 'STRING'
}, {
    'name': 'propertyName',
    'type': 'STRING'
}, {
    'name': 'propertyId',
    'type': 'STRING'
}, {
    'name': 'id',
    'type': 'STRING'
}, {
    'name': 'name',
    'type': 'STRING'
}, {
    'name': 'index',
    'type': 'STRING'
}, {
    'name': 'scope',
    'type': 'STRING'
}, {
    'name': 'active',
    'type': 'STRING'
}, {
    'name': 'created',
    'type': 'TIMESTAMP'
}, {
    'name': 'updated',
    'type': 'TIMESTAMP'
}, {
    'name': 'date',
    'type': 'DATE'
}, {
    'name': 'selfLink',
    'type': 'STRING'
}, {
    'name': 'type',
    'type': 'STRING'
}, {
    'name': 'min_value',
    'type': 'INTEGER'
}, {
    'name': 'max_value',
    'type': 'INTEGER'
}]

VIEW_SCHEMA = [{
    'name': 'date',
    'type': 'DATE'
}, {
    'name': 'id',
    'type': 'STRING'
}, {
    'name': 'selfLink',
    'type': 'STRING'
}, {
    'name': 'accountId',
    'type': 'STRING'
}, {
    'name': 'webPropertyId',
    'type': 'STRING'
}, {
    'name': 'accountName',
    'type': 'STRING'
}, {
    'name': 'webPropertyName',
    'type': 'STRING'
}, {
    'name': 'name',
    'type': 'STRING'
}, {
    'name': 'currency',
    'type': 'STRING'
}, {
    'name': 'timezone',
    'type': 'STRING'
}, {
    'name': 'websiteUrl',
    'type': 'STRING'
}, {
    'name': 'defaultPage',
    'type': 'STRING'
}, {
    'name': 'excludeQueryParameters',
    'type': 'STRING'
}, {
    'name': 'siteSearchQueryParameters',
    'type': 'STRING'
}, {
    'name': 'stripSiteSearchQueryParameters',
    'type': 'BOOLEAN'
}, {
    'name': 'siteSearchCategoryParameters',
    'type': 'STRING'
}, {
    'name': 'stripSiteSearchCategoryParameters',
    'type': 'BOOLEAN'
}, {
    'name': 'type',
    'type': 'STRING'
}, {
    'name': 'created',
    'type': 'TIMESTAMP'
}, {
    'name': 'updated',
    'type': 'TIMESTAMP'
}, {
    'name': 'eCommerceTracking',
    'type': 'BOOLEAN'
}, {
    'name': 'enhancedECommerceTracking',
    'type': 'BOOLEAN'
}, {
    'name': 'botFilteringEnabled',
    'type': 'BOOLEAN'
}, {
    'name': 'starred',
    'type': 'BOOLEAN'
}]

GOAL_SCHEMA = [{
    'name': 'date',
    'type': 'DATE'
}, {
    'name': 'id',
    'type': 'STRING'
}, {
    'name': 'accountId',
    'type': 'STRING'
}, {
    'name': 'webPropertyId',
    'type': 'STRING'
}, {
    'name': 'internalWebPropertyId',
    'type': 'STRING'
}, {
    'name': 'profileId',
    'type': 'STRING'
}, {
    'name': 'name',
    'type': 'STRING'
}, {
    'name': 'accountName',
    'type': 'STRING'
}, {
    'name': 'webPropertyName',
    'type': 'STRING'
}, {
    'name': 'profileName',
    'type': 'STRING'
}, {
    'name': 'value',
    'type': 'FLOAT'
}, {
    'name': 'active',
    'type': 'BOOLEAN'
}, {
    'name': 'type',
    'type': 'STRING'
}, {
    'name': 'created',
    'type': 'TIMESTAMP'
}, {
    'name': 'updated',
    'type': 'TIMESTAMP'
}, {
    'name':
        'urlDestinationDetails',
    'type':
        'RECORD',
    'fields': [{
        'name': 'url',
        'type': 'STRING'
    }, {
        'name': 'caseSensitive',
        'type': 'BOOLEAN'
    }, {
        'name': 'matchType',
        'type': 'STRING'
    }, {
        'name': 'firstStepRequired',
        'type': 'BOOLEAN'
    }, {
        'name':
            'steps',
        'type':
            'RECORD',
        'mode':
            'REPEATED',
        'fields': [{
            'name': 'number',
            'type': 'INTEGER'
        }, {
            'name': 'name',
            'type': 'STRING'
        }, {
            'name': 'url',
            'type': 'STRING'
        }]
    }]
}, {
    'name':
        'visitTimeOnSiteDetails',
    'type':
        'RECORD',
    'fields': [{
        'name': 'comparisonType',
        'type': 'STRING'
    }, {
        'name': 'comparisonValue',
        'type': 'STRING'
    }]
}, {
    'name':
        'visitNumPagesDetails',
    'type':
        'RECORD',
    'fields': [{
        'name': 'comparisonType',
        'type': 'STRING'
    }, {
        'name': 'comparisonValue',
        'type': 'STRING'
    }]
}, {
    'name':
        'eventDetails',
    'type':
        'RECORD',
    'fields': [{
        'name': 'useEventValue',
        'type': 'BOOLEAN'
    }, {
        'name':
            'eventConditions',
        'mode':
            'REPEATED',
        'type':
            'RECORD',
        'fields': [{
            'name': 'type',
            'type': 'STRING'
        }, {
            'name': 'matchType',
            'type': 'STRING'
        }, {
            'name': 'expression',
            'type': 'STRING'
        }, {
            'name': 'comparisonType',
            'type': 'STRING'
        }, {
            'name': 'comparisonValue',
            'type': 'STRING'
        }]
    }]
}]

GOOGLE_ADS_LINK_SCHEMA = [{
    'name': 'date',
    'type': 'DATE'
}, {
    'name': 'id',
    'type': 'STRING'
}, {
    'name': 'kind',
    'type': 'STRING'
}, {
    'name': 'selfLink',
    'type': 'STRING'
}, {
    'name':
        'entity',
    'type':
        'RECORD',
    'fields': [{
        'name':
            'webPropertyRef',
        'type':
            'RECORD',
        'fields': [{
            'name': 'id',
            'type': 'STRING'
        }, {
            'name': 'kind',
            'type': 'STRING'
        }, {
            'name': 'href',
            'type': 'STRING'
        }, {
            'name': 'accountId',
            'type': 'STRING'
        }, {
            'name': 'internalWebPropertyId',
            'type': 'STRING'
        }, {
            'name': 'name',
            'type': 'STRING'
        }]
    }]
}, {
    'name':
        'adWordsAccounts',
    'type':
        'RECORD',
    'mode':
        'REPEATED',
    'fields': [{
        'name': 'kind',
        'type': 'STRING'
    }, {
        'name': 'customerId',
        'type': 'STRING'
    }, {
        'name': 'autoTaggingEnabled',
        'type': 'BOOLEAN'
    }]
}, {
    'name': 'name',
    'type': 'STRING'
}, {
    'name': 'profileIds',
    'type': 'RECORD',
    'mode': 'REPEATED',
    'fields': [{
        'name': 'id',
        'type': 'STRING'
    }]
}]

REMARKETING_AUDIENCE_SCHEMA = [{
    'name': 'date',
    'type': 'DATE'
}, {
    'name': 'kind',
    'type': 'STRING'
}, {
    'name': 'id',
    'type': 'STRING'
}, {
    'name': 'accountId',
    'type': 'STRING'
}, {
    'name': 'webPropertyId',
    'type': 'STRING'
}, {
    'name': 'webPropertyName',
    'type': 'STRING'
}, {
    'name': 'accountName',
    'type': 'STRING'
}, {
    'name': 'internalWebPropertyId',
    'type': 'STRING'
}, {
    'name': 'created',
    'type': 'TIMESTAMP'
}, {
    'name': 'updated',
    'type': 'TIMESTAMP'
}, {
    'name': 'name',
    'type': 'STRING'
}, {
    'name': 'description',
    'type': 'STRING'
}, {
    'name':
        'linkedAdAccounts',
    'mode':
        'REPEATED',
    'type':
        'RECORD',
    'fields': [{
        'name': 'kind',
        'type': 'STRING'
    }, {
        'name': 'id',
        'type': 'STRING'
    }, {
        'name': 'accountId',
        'type': 'STRING'
    }, {
        'name': 'webPropertyId',
        'type': 'STRING'
    }, {
        'name': 'internalWebPropertyId',
        'type': 'STRING'
    }, {
        'name': 'remarketingAudienceId',
        'type': 'STRING'
    }, {
        'name': 'linkedAccountId',
        'type': 'STRING'
    }, {
        'name': 'type',
        'type': 'STRING'
    }, {
        'name': 'status',
        'type': 'STRING'
    }, {
        'name': 'eligibleForSearch',
        'type': 'BOOLEAN'
    }]
}, {
    'name': 'linkedViews',
    'type': 'RECORD',
    'mode': 'REPEATED',
    'fields': [{
        'name': 'id',
        'type': 'STRING'
    }]
}, {
    'name': 'audienceType',
    'type': 'STRING'
}, {
    'name':
        'audienceDefinition',
    'type':
        'RECORD',
    'fields': [{
        'name':
            'includeConditions',
        'type':
            'RECORD',
        'fields': [{
            'name': 'kind',
            'type': 'STRING'
        }, {
            'name': 'isSmartList',
            'type': 'BOOLEAN'
        }, {
            'name': 'segment',
            'type': 'STRING'
        }, {
            'name': 'membershipDurationDays',
            'type': 'INTEGER'
        }, {
            'name': 'daysToLookBack',
            'type': 'INTEGER'
        }]
    }]
}, {
    'name':
        'stateBasedAudienceDefinition',
    'type':
        'RECORD',
    'fields': [{
        'name':
            'includeConditions',
        'type':
            'RECORD',
        'fields': [{
            'name': 'kind',
            'type': 'STRING'
        }, {
            'name': 'isSmartList',
            'type': 'BOOLEAN'
        }, {
            'name': 'segment',
            'type': 'STRING'
        }, {
            'name': 'membershipDurationDays',
            'type': 'INTEGER'
        }, {
            'name': 'daysToLookBack',
            'type': 'INTEGER'
        }]
    }, {
        'name':
            'excludeConditions',
        'type':
            'RECORD',
        'fields': [{
            'name': 'segment',
            'type': 'STRING'
        }, {
            'name': 'exclusionDuration',
            'type': 'STRING'
        }]
    }]
}]

ACCOUNT_SUMMARIES_SCHEMA = [{
    'name': 'date',
    'type': 'DATE'
}, {
    'name': 'id',
    'type': 'STRING'
}, {
    'name': 'kind',
    'type': 'STRING'
}, {
    'name': 'name',
    'type': 'STRING'
}, {
    'name': 'starred',
    'type': 'BOOLEAN'
}, {
    'name':
        'webProperties',
    'type':
        'RECORD',
    'mode':
        'REPEATED',
    'fields': [{
        'name': 'kind',
        'type': 'STRING'
    }, {
        'name': 'id',
        'type': 'STRING'
    }, {
        'name': 'name',
        'type': 'STRING'
    }, {
        'name': 'internalWebPropertyId',
        'type': 'STRING'
    }, {
        'name': 'level',
        'type': 'STRING'
    }, {
        'name': 'websiteUrl',
        'type': 'STRING'
    }, {
        'name': 'starred',
        'type': 'BOOLEAN'
    }, {
        'name':
            'profiles',
        'type':
            'RECORD',
        'mode':
            'REPEATED',
        'fields': [
            {
                'name': 'kind',
                'type': 'STRING'
            },
            {
                'name': 'id',
                'type': 'STRING'
            },
            {
                'name': 'name',
                'type': 'STRING'
            },
            {
                'name': 'type',
                'type': 'STRING'
            },
            {
                'name': 'starred',
                'type': 'BOOLEAN'
            },
        ]
    }]
}]



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

Campaign_Schema = [
    {
        'name':
            'common_data',
        'type':
            'RECORD',
        'mode':
            'NULLABLE',
        'fields': [
            {
                'name': 'id',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'name',
                'type': 'STRING',
                'mode': 'NULLABLE',
            },
            {
                'name': 'active',
                'type': 'BOOLEAN',
                'mode': 'NULLABLE',
            },
            {
                'name': 'integration_code',
                'type': 'STRING',
                'mode': 'NULLABLE',
            },
        ]
    },
    {
        'name': 'advertiser_id',
        'type': 'INTEGER',
        'mode': 'NULLABLE',
    },
    {
        'name':
            'budget',
        'type':
            'RECORD',
        'mode':
            'NULLABLE',
        'fields': [
            {
                'name': 'start_time_usec',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'end_time_usec',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'max_impressions',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'max_spend_advertiser_micros',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'pacing_type',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'pacing_max_impressions',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'pacing_max_spend_advertiser_micros',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'pacing_distribution',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
        ]
    },
    {
        'name':
            'frequency_cap',
        'type':
            'RECORD',
        'mode':
            'NULLABLE',
        'fields': [
            {
                'name': 'max_impressions',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'time_unit',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'time_range',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
        ]
    },
    {
        'name':
            'default_target_list',
        'type':
            'RECORD',
        'mode':
            'NULLABLE',
        'fields': [
            {
                'name':
                    'inventory_sources',
                'type':
                    'RECORD',
                'mode':
                    'REPEATED',
                'fields': [
                    {
                        'name': 'criteria_id',
                        'type': 'INTEGER',
                        'mode': 'NULLABLE',
                    },
                    {
                        'name': 'excluded',
                        'type': 'BOOLEAN',
                        'mode': 'NULLABLE',
                    },
                ]
            },
            {
                'name':
                    'geo_locations',
                'type':
                    'RECORD',
                'mode':
                    'REPEATED',
                'fields': [
                    {
                        'name': 'criteria_id',
                        'type': 'INTEGER',
                        'mode': 'NULLABLE',
                    },
                    {
                        'name': 'excluded',
                        'type': 'BOOLEAN',
                        'mode': 'NULLABLE',
                    },
                ]
            },
            {
                'name':
                    'ad_position',
                'type':
                    'RECORD',
                'mode':
                    'NULLABLE',
                'fields': [
                    {
                        'name': 'criteria_id',
                        'type': 'INTEGER',
                        'mode': 'NULLABLE',
                    },
                    {
                        'name': 'excluded',
                        'type': 'BOOLEAN',
                        'mode': 'NULLABLE',
                    },
                ]
            },
            {
                'name':
                    'net_speed',
                'type':
                    'RECORD',
                'mode':
                    'NULLABLE',
                'fields': [
                    {
                        'name': 'criteria_id',
                        'type': 'INTEGER',
                        'mode': 'NULLABLE',
                    },
                    {
                        'name': 'excluded',
                        'type': 'BOOLEAN',
                        'mode': 'NULLABLE',
                    },
                ]
            },
            {
                'name':
                    'browsers',
                'type':
                    'RECORD',
                'mode':
                    'NULLABLE',
                'fields': [
                    {
                        'name':
                            'union',
                        'type':
                            'RECORD',
                        'mode':
                            'REPEATED',
                        'fields': [
                            {
                                'name': 'criteria_id',
                                'type': 'INTEGER',
                                'mode': 'NULLABLE',
                            },
                            {
                                'name': 'parameter',
                                'type': 'STRING',
                                'mode': 'NULLABLE',
                            },
                        ]
                    },
                    {
                        'name': 'excluded',
                        'type': 'BOOLEAN',
                        'mode': 'NULLABLE',
                    },
                ]
            },
            {
                'name':
                    'device_criteria',
                'type':
                    'RECORD',
                'mode':
                    'REPEATED',
                'fields': [
                    {
                        'name': 'criteria_id',
                        'type': 'INTEGER',
                        'mode': 'NULLABLE',
                    },
                    {
                        'name': 'excluded',
                        'type': 'BOOLEAN',
                        'mode': 'NULLABLE',
                    },
                ]
            },
            {
                'name':
                    'languages',
                'type':
                    'RECORD',
                'mode':
                    'NULLABLE',
                'fields': [
                    {
                        'name':
                            'union',
                        'type':
                            'RECORD',
                        'mode':
                            'REPEATED',
                        'fields': [
                            {
                                'name': 'criteria_id',
                                'type': 'INTEGER',
                                'mode': 'NULLABLE',
                            },
                            {
                                'name': 'parameter',
                                'type': 'STRING',
                                'mode': 'NULLABLE',
                            },
                        ]
                    },
                    {
                        'name': 'excluded',
                        'type': 'BOOLEAN',
                        'mode': 'NULLABLE',
                    },
                ]
            },
            {
                'name':
                    'day_parting',
                'type':
                    'RECORD',
                'mode':
                    'NULLABLE',
                'fields': [
                    {
                        'name':
                            'union',
                        'type':
                            'RECORD',
                        'mode':
                            'REPEATED',
                        'fields': [
                            {
                                'name': 'criteria_id',
                                'type': 'INTEGER',
                                'mode': 'NULLABLE',
                            },
                            {
                                'name': 'parameter',
                                'type': 'STRING',
                                'mode': 'NULLABLE',
                            },
                        ]
                    },
                    {
                        'name': 'excluded',
                        'type': 'BOOLEAN',
                        'mode': 'NULLABLE',
                    },
                ]
            },
            {
                'name':
                    'audience_intersect',
                'type':
                    'RECORD',
                'mode':
                    'REPEATED',
                'fields': [
                    {
                        'name':
                            'union',
                        'type':
                            'RECORD',
                        'mode':
                            'REPEATED',
                        'fields': [
                            {
                                'name': 'criteria_id',
                                'type': 'INTEGER',
                                'mode': 'NULLABLE',
                            },
                            {
                                'name': 'parameter',
                                'type': 'STRING',
                                'mode': 'NULLABLE',
                            },
                        ]
                    },
                    {
                        'name': 'excluded',
                        'type': 'BOOLEAN',
                        'mode': 'NULLABLE',
                    },
                ]
            },
            {
                'name':
                    'keywords',
                'type':
                    'RECORD',
                'mode':
                    'REPEATED',
                'fields': [
                    {
                        'name': 'criteria_id',
                        'type': 'INTEGER',
                        'mode': 'NULLABLE',
                    },
                    {
                        'name': 'parameter',
                        'type': 'STRING',
                        'mode': 'NULLABLE',
                    },
                    {
                        'name': 'excluded',
                        'type': 'BOOLEAN',
                        'mode': 'NULLABLE',
                    },
                ]
            },
            {
                'name': 'kct_include_uncrawled_sites',
                'type': 'BOOLEAN',
                'mode': 'NULLABLE',
            },
            {
                'name':
                    'page_categories',
                'type':
                    'RECORD',
                'mode':
                    'REPEATED',
                'fields': [
                    {
                        'name': 'criteria_id',
                        'type': 'INTEGER',
                        'mode': 'NULLABLE',
                    },
                    {
                        'name': 'excluded',
                        'type': 'BOOLEAN',
                        'mode': 'NULLABLE',
                    },
                ]
            },
            {
                'name':
                    'universal_channels',
                'type':
                    'RECORD',
                'mode':
                    'REPEATED',
                'fields': [
                    {
                        'name': 'criteria_id',
                        'type': 'INTEGER',
                        'mode': 'NULLABLE',
                    },
                    {
                        'name': 'excluded',
                        'type': 'BOOLEAN',
                        'mode': 'NULLABLE',
                    },
                ]
            },
            {
                'name':
                    'sites',
                'type':
                    'RECORD',
                'mode':
                    'REPEATED',
                'fields': [
                    {
                        'name': 'criteria_id',
                        'type': 'INTEGER',
                        'mode': 'NULLABLE',
                    },
                    {
                        'name': 'parameter',
                        'type': 'STRING',
                        'mode': 'NULLABLE',
                    },
                    {
                        'name': 'excluded',
                        'type': 'BOOLEAN',
                        'mode': 'NULLABLE',
                    },
                ]
            },
            {
                'name':
                    'brand_safety',
                'type':
                    'RECORD',
                'mode':
                    'NULLABLE',
                'fields': [
                    {
                        'name': 'criteria_id',
                        'type': 'INTEGER',
                        'mode': 'NULLABLE',
                    },
                    {
                        'name': 'excluded',
                        'type': 'BOOLEAN',
                        'mode': 'NULLABLE',
                    },
                ]
            },
        ]
    },
    {
        'name': 'uses_video_creatives',
        'type': 'BOOLEAN',
        'mode': 'NULLABLE',
    },
    {
        'name': 'uses_display_creatives',
        'type': 'BOOLEAN',
        'mode': 'NULLABLE',
    },
    {
        'name': 'uses_audio_creatives',
        'type': 'BOOLEAN',
        'mode': 'NULLABLE',
    },
    {
        'name': 'objective',
        'type': 'INTEGER',
        'mode': 'NULLABLE',
    },
    {
        'name': 'objective_description',
        'type': 'STRING',
        'mode': 'NULLABLE',
    },
    {
        'name': 'metric',
        'type': 'INTEGER',
        'mode': 'NULLABLE',
    },
    {
        'name': 'metric_amount_micros',
        'type': 'INTEGER',
        'mode': 'NULLABLE',
    },
]



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import re

from starthinker.util.project import project
from starthinker.util.bigquery import table_create
from starthinker.util.google_api import API_DCM
from starthinker.util.data import put_rows, get_rows
from starthinker.util.dcm import get_profile_for_api, id_to_timezone
from starthinker.util.regexp import epoch_to_datetime

PROFILE_CAMPAIGNS = []
PROFILE_SITES = []
PROFILE_ROLES = []
PROFILE_ADVERTISERS = []
REPORT_DELIVERIES = []
SITE_CONTACTS = []

ACCOUNTS_SCHEMA = [
    {
        'name': 'accountId',
        'type': 'INTEGER'
    },
    {
        'name': 'name',
        'type': 'STRING'
    },
    {
        'name': 'active',
        'type': 'BOOLEAN'
    },
    {
        'name': 'description',
        'type': 'STRING'
    },
    {
        'name': 'timezone',
        'type': 'STRING'
    },
    {
        'name': 'currencyId',
        'type': 'INTEGER'
    },
    {
        'name': 'countryId',
        'type': 'INTEGER'
    },
    {
        'name': 'locale',
        'type': 'STRING'
    },
    {
        'name': 'nielsenOcrEnabled',
        'type': 'BOOLEAN'
    },
    {
        'name': 'shareReportsWithTwitter',
        'type': 'BOOLEAN'
    },
]

PROFILES_SCHEMA = [
    {
        'name': 'profileId',
        'type': 'INTEGER'
    },
    {
        'name': 'accountId',
        'type': 'INTEGER'
    },
    {
        'name': 'subaccountId',
        'type': 'INTEGER'
    },
    {
        'name': 'name',
        'type': 'STRING'
    },
    {
        'name': 'email',
        'type': 'STRING'
    },
    {
        'name': 'locale',
        'type': 'STRING'
    },
    {
        'name': 'userRoleId',
        'type': 'INTEGER'
    },
    {
        'name': 'userAccessType',
        'type': 'STRING'
    },
    {
        'name': 'active',
        'type': 'BOOLEAN'
    },
    {
        'name': 'comments',
        'type': 'STRING'
    },
    {
        'name': 'traffickerType',
        'type': 'STRING'
    },
]

PROFILE_CAMPAIGNS_SCHEMA = [
    {
        'name': 'profileId',
        'type': 'INTEGER'
    },
    {
        'name': 'accountId',
        'type': 'INTEGER'
    },
    {
        'name': 'subaccountId',
        'type': 'INTEGER'
    },
    {
        'name': 'campaignId',
        'type': 'INTEGER'
    },
]

PROFILE_SITES_SCHEMA = [
    {
        'name': 'profileId',
        'type': 'INTEGER'
    },
    {
        'name': 'accountId',
        'type': 'INTEGER'
    },
    {
        'name': 'subaccountId',
        'type': 'INTEGER'
    },
    {
        'name': 'siteId',
        'type': 'INTEGER'
    },
]

PROFILE_ROLES_SCHEMA = [
    {
        'name': 'profileId',
        'type': 'INTEGER'
    },
    {
        'name': 'accountId',
        'type': 'INTEGER'
    },
    {
        'name': 'subaccountId',
        'type': 'INTEGER'
    },
    {
        'name': 'roleId',
        'type': 'INTEGER'
    },
]

PROFILE_ADVERTISERS_SCHEMA = [
    {
        'name': 'profileId',
        'type': 'INTEGER'
    },
    {
        'name': 'accountId',
        'type': 'INTEGER'
    },
    {
        'name': 'subaccountId',
        'type': 'INTEGER'
    },
    {
        'name': 'advertiserId',
        'type': 'INTEGER'
    },
]

ADVERTISERS_SCHEMA = [
    {
        'name': 'accountId',
        'type': 'INTEGER'
    },
    {
        'name': 'subaccountId',
        'type': 'INTEGER'
    },
    {
        'name': 'advertiserId',
        'type': 'INTEGER'
    },
    {
        'name': 'advertiserGroupId',
        'type': 'INTEGER'
    },
    {
        'name': 'name',
        'type': 'STRING'
    },
    {
        'name': 'status',
        'type': 'STRING'
    },
    {
        'name': 'defaultEmail',
        'type': 'STRING'
    },
    {
        'name': 'clickThroughUrlSuffix',
        'type': 'STRING'
    },
    {
        'name': 'defaultClickThroughEventTagId',
        'type': 'INTEGER'
    },
    {
        'name': 'suspended',
        'type': 'BOOLEAN'
    },
    {
        'name': 'floodlightConfigurationId',
        'type': 'INTEGER'
    },
    {
        'name': 'originalFloodlightConfigurationId',
        'type': 'INTEGER'
    },
]

CAMPAIGNS_SCHEMA = [
    {
        'name': 'accountId',
        'type': 'INTEGER'
    },
    {
        'name': 'subaccountId',
        'type': 'INTEGER'
    },
    {
        'name': 'advertiserId',
        'type': 'INTEGER'
    },
    {
        'name': 'advertiserGroupId',
        'type': 'INTEGER'
    },
    {
        'name': 'campaignId',
        'type': 'INTEGER'
    },
    {
        'name': 'name',
        'type': 'STRING'
    },
    {
        'name': 'archived',
        'type': 'BOOLEAN'
    },
    {
        'name': 'startDate',
        'type': 'DATE'
    },
    {
        'name': 'endDate',
        'type': 'DATE'
    },
    {
        'name': 'comment',
        'type': 'STRING'
    },
    {
        'name': 'createInfo_time',
        'type': 'STRING'
    },
    {
        'name': 'lastModifiedInfo_time',
        'type': 'TIMESTAMP'
    },
    {
        'name': 'externalId',
        'type': 'STRING'
    },
    {
        'name': 'defaultLandingPageId',
        'type': 'STRING'
    },
    {
        'name': 'adBlockingConfiguration_enabled',
        'type': 'BOOLEAN'
    },
    {
        'name': 'adBlockingConfiguration_overrideClickThroughUrl',
        'type': 'STRING'
    },
    {
        'name': 'adBlockingConfiguration_clickThroughUrl',
        'type': 'STRING'
    },
    {
        'name': 'adBlockingConfiguration_creativeBundleId',
        'type': 'STRING'
    },
    {
        'name': 'nielsenOcrEnabled',
        'type': 'BOOLEAN'
    },
]

SITES_SCHEMA = [
    {
        'name': 'accountId',
        'type': 'INTEGER'
    },
    {
        'name': 'subaccountId',
        'type': 'INTEGER'
    },
    {
        'name': 'directorySiteId',
        'type': 'INTEGER'
    },
    {
        'name': 'siteId',
        'type': 'INTEGER'
    },
    {
        'name': 'name',
        'type': 'STRING'
    },
    {
        'name': 'keyName',
        'type': 'STRING'
    },
    {
        'name': 'approved',
        'type': 'BOOLEAN'
    },
    {
        'name': 'orientation',
        'type': 'STRING'
    },
    {
        'name': 'siteSettings_disableNewCookie',
        'type': 'BOOLEAN'
    },
    {
        'name': 'siteSettings_activeViewOptOut',
        'type': 'BOOLEAN'
    },
    {
        'name': 'siteSettings_adBlockingOptOut',
        'type': 'BOOLEAN'
    },
    {
        'name': 'siteSettings_videoActiveViewOptOutTemplate',
        'type': 'BOOLEAN'
    },
    {
        'name': 'siteSettings_vpaidAdapterChoiceTemplate',
        'type': 'STRING'
    },
    {
        'name': 'siteSettings_tagSetting_additionalKeyValues',
        'type': 'STRING'
    },
    {
        'name': 'siteSettings_tagSetting_includeClickTracking',
        'type': 'BOOLEAN'
    },
    {
        'name': 'siteSettings_tagSetting_includeClickThroughUrls',
        'type': 'BOOLEAN'
    },
    {
        'name': 'siteSettings_tagSetting_keywordOption',
        'type': 'STRING'
    },
]

SITE_CONTACTS_SCHEMA = [
    {
        'name': 'accountId',
        'type': 'INTEGER'
    },
    {
        'name': 'subaccountId',
        'type': 'INTEGER'
    },
    {
        'name': 'directorySiteId',
        'type': 'INTEGER'
    },
    {
        'name': 'siteId',
        'type': 'INTEGER'
    },
    {
        'name': 'contactId',
        'type': 'INTEGER'
    },
    {
        'name': 'email',
        'type': 'STRING'
    },
    {
        'name': 'firstName',
        'type': 'STRING'
    },
    {
        'name': 'lastName',
        'type': 'STRING'
    },
    {
        'name': 'title',
        'type': 'STRING'
    },
    {
        'name': 'address',
        'type': 'STRING'
    },
    {
        'name': 'phone',
        'type': 'STRING'
    },
    {
        'name': 'contactType',
        'type': 'STRING'
    },
]

ROLES_SCHEMA = [
    {
        'name': 'accountId',
        'type': 'INTEGER'
    },
    {
        'name': 'subaccountId',
        'type': 'INTEGER'
    },
    {
        'name': 'roleId',
        'type': 'INTEGER'
    },
    {
        'name': 'role_name',
        'type': 'STRING'
    },
    {
        'name': 'role_defaultUserRole',
        'type': 'BOOLEAN'
    },
    {
        'name': 'permission_name',
        'type': 'STRING'
    },
    {
        'name': 'permission_availability',
        'type': 'STRING'
    },
]

SUBACCOUNTS_SCHEMA = [
    {
        'name': 'accountId',
        'type': 'INTEGER'
    },
    {
        'name': 'subaccountId',
        'type': 'INTEGER'
    },
    {
        'name': 'name',
        'type': 'STRING'
    },
]

REPORTS_SCHEMA = [
    {
        'name': 'profileId',
        'type': 'INTEGER'
    },
    {
        'name': 'accountId',
        'type': 'INTEGER'
    },
    {
        'name': 'subaccountId',
        'type': 'INTEGER'
    },
    {
        'name': 'reportId',
        'type': 'INTEGER'
    },
    {
        'name': 'name',
        'type': 'STRING'
    },
    {
        'name': 'type',
        'type': 'STRING'
    },
    {
        'name': 'format',
        'type': 'STRING'
    },
    {
        'name': 'lastModifiedTime',
        'type': 'TIMESTAMP'
    },
    {
        'name': 'criteria_startDate',
        'type': 'DATE'
    },
    {
        'name': 'criteria_endDate',
        'type': 'DATE'
    },
    {
        'name': 'criteria_relativeDateRange',
        'type': 'STRING'
    },
    {
        'name': 'schedule_active',
        'type': 'BOOLEAN'
    },
    {
        'name': 'schedule_startDate',
        'type': 'DATE'
    },
    {
        'name': 'schedule_expirationDate',
        'type': 'DATE'
    },
    {
        'name': 'schedule_runsOnDayOfMonth',
        'type': 'STRING'
    },
    {
        'name': 'schedule_repeats',
        'type': 'STRING'
    },
    {
        'name': 'schedule_every',
        'type': 'STRING'
    },
    {
        'name': 'schedule_repeatsOnWeekDays',
        'type': 'STRING'
    },
]

REPORT_DELIVERIES_SCHEMA = [
    {
        'name': 'profileId',
        'type': 'INTEGER'
    },
    {
        'name': 'accountId',
        'type': 'INTEGER'
    },
    {
        'name': 'subaccountId',
        'type': 'INTEGER'
    },
    {
        'name': 'reportId',
        'type': 'INTEGER'
    },
    {
        'name': 'emailOwner',
        'type': 'STRING'
    },
    {
        'name': 'emailOwnerDeliveryType',
        'type': 'STRING'
    },
    {
        'name': 'message',
        'type': 'STRING'
    },
    {
        'name': 'email',
        'type': 'STRING'
    },
    {
        'name': 'deliveryType',
        'type': 'STRING'
    },
]


def get_accounts(accounts):
  if project.verbose:
    print('DCM Accounts')

  for account_id in accounts:
    is_superuser, profile_id = get_profile_for_api(project.task['auth'],
                                                   account_id)
    kwargs = {'profileId': profile_id, 'id': account_id}
    account = API_DCM('user').accounts().get(**kwargs).execute()
    yield [
        account['id'],
        account['name'],
        account['active'],
        account.get('description', ''),
        id_to_timezone(
            account['reportsConfiguration']['reportGenerationTimeZoneId']),
        account.get('currencyId'),
        account.get('countryId'),
        account['locale'],
        account['nielsenOcrEnabled'],
        account['shareReportsWithTwitter'],
    ]


def get_profiles(accounts):
  if project.verbose:
    print('DCM Profiles')

  for account_id in accounts:
    is_superuser, profile_id = get_profile_for_api(project.task['auth'],
                                                   account_id)
    kwargs = {
        'profileId': profile_id,
        'accountId': account_id
    } if is_superuser else {
        'profileId': profile_id
    }
    for profile in API_DCM(
        'user', iterate=True,
        internal=is_superuser).accountUserProfiles().list(**kwargs).execute():
      if int(profile['accountId']) in accounts:

        for campaign in profile.get('campaignFilter', {}).get('objectIds', []):
          PROFILE_CAMPAIGNS.append([
              profile['id'],
              profile['accountId'],
              profile.get('subaccountId'),
              campaign,
          ])

        for site in profile.get('siteFilter', {}).get('objectIds', []):
          PROFILE_SITES.append([
              profile['id'],
              profile['accountId'],
              profile.get('subaccountId'),
              site,
          ])

        for role in profile.get('userRoleFilter', {}).get('objectIds', []):
          PROFILE_ROLES.append([
              profile['id'],
              profile['accountId'],
              profile.get('subaccountId'),
              role,
          ])

        for advertiser in profile.get('advertiserFilter',
                                      {}).get('objectIds', []):
          PROFILE_ADVERTISERS.append([
              profile['id'],
              profile['accountId'],
              profile.get('subaccountId'),
              advertiser,
          ])

        yield [
            profile['id'],
            profile['accountId'],
            profile.get('subaccountId'),
            profile['name'],
            profile['email'],
            profile.get('locale'),
            profile.get('userRoleId'),
            profile.get('userAccessType'),
            profile['active'],
            profile.get('comments', ''),
            profile.get('traffickerType'),
        ]


def get_subaccounts(accounts):

  if project.verbose:
    print('DCM SubAccounts')

  for account_id in accounts:
    is_superuser, profile_id = get_profile_for_api(project.task['auth'],
                                                   account_id)
    kwargs = {
        'profileId': profile_id,
        'accountId': account_id
    } if is_superuser else {
        'profileId': profile_id
    }
    for subaccount in API_DCM(
        'user', iterate=True,
        internal=is_superuser).subaccounts().list(**kwargs).execute():
      if int(subaccount['accountId']) in accounts:
        yield [
            subaccount['accountId'],
            subaccount['id'],
            subaccount['name'],
        ]


def get_advertisers(accounts):

  if project.verbose:
    print('DCM Advertisers')

  for account_id in accounts:
    is_superuser, profile_id = get_profile_for_api(project.task['auth'],
                                                   account_id)
    kwargs = {
        'profileId': profile_id,
        'accountId': account_id
    } if is_superuser else {
        'profileId': profile_id
    }
    for advertiser in API_DCM(
        'user', iterate=True,
        internal=is_superuser).advertisers().list(**kwargs).execute():
      if int(advertiser['accountId']) in accounts:
        yield [
            advertiser['accountId'],
            advertiser.get('subaccountId'),
            advertiser['id'],
            advertiser.get('advertiserGroupId'),
            advertiser['name'],
            advertiser['status'],
            advertiser.get('defaultEmail'),
            advertiser.get('clickThroughUrlSuffix'),
            advertiser.get('defaultClickThroughEventTagId'),
            advertiser['suspended'],
            advertiser['floodlightConfigurationId'],
            advertiser['originalFloodlightConfigurationId'],
        ]


def get_campaigns(accounts):

  if project.verbose:
    print('DCM Campaigns')

  for account_id in accounts:
    is_superuser, profile_id = get_profile_for_api(project.task['auth'],
                                                   account_id)
    kwargs = {
        'profileId': profile_id,
        'accountId': account_id
    } if is_superuser else {
        'profileId': profile_id
    }
    for campaign in API_DCM(
        'user', iterate=True,
        internal=is_superuser).campaigns().list(**kwargs).execute():
      if int(campaign['accountId']) in accounts:

        yield [
            campaign['accountId'],
            campaign.get('subaccountId'),
            campaign['advertiserId'],
            campaign.get('advertiserGroupId'),
            campaign['id'],
            campaign['name'],
            campaign['archived'],
            campaign['startDate'],
            campaign['endDate'],
            campaign.get('comment', ''),
            epoch_to_datetime(campaign['createInfo']['time'], 1000),
            epoch_to_datetime(campaign['lastModifiedInfo']['time'], 1000),
            campaign.get('externalId'),
            campaign['defaultLandingPageId'],
            campaign['adBlockingConfiguration']['enabled'],
            campaign['adBlockingConfiguration']['overrideClickThroughUrl'],
            campaign['adBlockingConfiguration'].get('clickThroughUrl'),
            campaign['adBlockingConfiguration'].get('creativeBundleId'),
            campaign['nielsenOcrEnabled'],
        ]


def get_sites(accounts):

  if project.verbose:
    print('DCM Sites')

  for account_id in accounts:
    is_superuser, profile_id = get_profile_for_api(project.task['auth'],
                                                   account_id)
    kwargs = {
        'profileId': profile_id,
        'accountId': account_id
    } if is_superuser else {
        'profileId': profile_id
    }
    for site in API_DCM(
        'user', iterate=True,
        internal=is_superuser).sites().list(**kwargs).execute():
      if int(site['accountId']) in accounts:

        for contact in site.get('siteContacts', []):
          SITE_CONTACTS.append([
              site['accountId'],
              site.get('subaccountId'),
              site.get('directorySiteId'),
              site['id'],
              contact['id'],
              contact['email'],
              contact.get('firstName', ''),
              contact.get('lastName', ''),
              contact.get('title', ''),
              contact.get('address', ''),
              contact.get('phone', ''),
              contact['contactType'],
          ])

        yield [
            site['accountId'],
            site.get('subaccountId'),
            site.get('directorySiteId'),
            site['id'],
            site['name'],
            site['keyName'],
            site['approved'],
            site.get('orientation'),
            site['siteSettings'].get('disableNewCookie'),
            site['siteSettings'].get('activeViewOptOut'),
            site['siteSettings'].get('adBlockingOptOut'),
            site['siteSettings'].get('videoActiveViewOptOutTemplate'),
            site['siteSettings'].get('vpaidAdapterChoiceTemplate'),
            site['siteSettings'].get('tagSetting',
                                     {}).get('additionalKeyValues'),
            site['siteSettings'].get('tagSetting',
                                     {}).get('includeClickTracking'),
            site['siteSettings'].get('tagSetting',
                                     {}).get('includeClickThroughUrls'),
            site['siteSettings'].get('tagSetting', {}).get('keywordOption'),
        ]


def get_roles(accounts):
  if project.verbose:
    print('DCM Roles')

  for account_id in accounts:
    is_superuser, profile_id = get_profile_for_api(project.task['auth'],
                                                   account_id)
    kwargs = {
        'profileId': profile_id,
        'accountId': account_id
    } if is_superuser else {
        'profileId': profile_id
    }
    for role in API_DCM(
        'user', iterate=True,
        internal=is_superuser).userRoles().list(**kwargs).execute():
      if int(role['accountId']) in accounts:
        if 'permissions' in role:
          for permission in role['permissions']:
            yield [
                role['accountId'],
                role.get('subaccountId'),
                role['id'],
                role['name'],
                role['defaultUserRole'],
                permission['name'],
                permission['availability'],
            ]
        else:
          yield [
              role['accountId'],
              role.get('subaccountId'),
              role['id'],
              role['name'],
              role['defaultUserRole'],
              None,
              None,
          ]


def get_reports(accounts):

  if project.verbose:
    print('DCM Reports')

  for account_id in accounts:
    is_superuser, profile_id = get_profile_for_api(project.task['auth'],
                                                   account_id)
    kwargs = {
        'profileId': profile_id,
        'accountId': account_id,
        'scope': 'ALL'
    } if is_superuser else {
        'profileId': profile_id,
        'scope': 'ALL'
    }
    for report in API_DCM(
        'user', iterate=True,
        internal=is_superuser).reports().list(**kwargs).execute():
      if int(report['accountId']) in accounts:

        for delivery in report.get('delivery', {}).get('recipients', []):
          REPORT_DELIVERIES.append((
              report['ownerProfileId'],
              report['accountId'],
              report.get('subaccountId'),
              report['id'],
              report['delivery']['emailOwner'],
              report['delivery'].get('emailOwnerDeliveryType'),
              report['delivery'].get('message', ''),
              delivery['email'],
              delivery['deliveryType'],
          ))

        yield [
            report['ownerProfileId'],
            report['accountId'],
            report.get('subaccountId'),
            report['id'],
            report['name'],
            report['type'],
            report.get('format'),
            epoch_to_datetime(report.get('lastModifiedTime'), 1000),
            report.get('criteria', {}).get('startDate'),
            report.get('criteria', {}).get('endDate'),
            report.get('criteria', {}).get('relativeDateRange'),
            report.get('schedule', {}).get('active'),
            report.get('schedule', {}).get('startDate'),
            report.get('schedule', {}).get('expirationDate'),
            report.get('schedule', {}).get('runsOnDayOfMonth'),
            report.get('schedule', {}).get('repeats'),
            report.get('schedule', {}).get('every'),
            ','.join(report.get('schedule', {}).get('repeatsOnWeekDays', [])),
        ]


def put_json(kind, schema):

  return {
      'bigquery': {
          'dataset': project.task['out']['dataset'],
          'table': kind,
          'schema': schema,
          'skip_rows': 0,
          'format': 'CSV'
      }
  }


@project.from_parameters
def barnacle():
  if project.verbose:
    print('BARNACLE')

  accounts = set(get_rows('user', project.task['accounts']))

  # Accounts
  rows = get_accounts(accounts)
  put_rows(project.task['out']['auth'], put_json('CM_Accounts',
                                                 ACCOUNTS_SCHEMA), rows)

  # Profiles
  rows = get_profiles(accounts)
  put_rows(project.task['out']['auth'], put_json('CM_Profiles',
                                                 PROFILES_SCHEMA), rows)

  # Profiles Campaigns
  if project.verbose:
    print('DCM Profile Campaigns')
  put_rows(project.task['out']['auth'],
           put_json('CM_Profile_Campaigns', PROFILE_CAMPAIGNS_SCHEMA),
           PROFILE_CAMPAIGNS)

  # Profiles Sites
  if project.verbose:
    print('DCM Profile Sites')
  put_rows(project.task['out']['auth'],
           put_json('CM_Profile_Sites', PROFILE_SITES_SCHEMA), PROFILE_SITES)

  # Profiles Roles
  if project.verbose:
    print('DCM Profile Roles')
  put_rows(project.task['out']['auth'],
           put_json('CM_Profile_Roles', PROFILE_ROLES_SCHEMA), PROFILE_ROLES)

  # Profiles Advertisers
  if project.verbose:
    print('DCM Profile Advertisers')
  put_rows(project.task['out']['auth'],
           put_json('CM_Profile_Advertisers', PROFILE_ADVERTISERS_SCHEMA),
           PROFILE_ADVERTISERS)

  # Subaccounts
  rows = get_subaccounts(accounts)
  put_rows(project.task['out']['auth'],
           put_json('CM_SubAccounts', SUBACCOUNTS_SCHEMA), rows)

  # Advertisers
  rows = get_advertisers(accounts)
  put_rows(project.task['out']['auth'],
           put_json('CM_Advertisers', ADVERTISERS_SCHEMA), rows)

  # Campaigns
  rows = get_campaigns(accounts)
  put_rows(project.task['out']['auth'],
           put_json('CM_Campaigns', CAMPAIGNS_SCHEMA), rows)

  # Sites
  rows = get_sites(accounts)
  put_rows(project.task['out']['auth'], put_json('CM_Sites', SITES_SCHEMA),
           rows)

  # Sites Contacts
  if project.verbose:
    print('DCM Site Contacts')
  put_rows(project.task['out']['auth'],
           put_json('CM_Site_Contacts', SITE_CONTACTS_SCHEMA), SITE_CONTACTS)

  # Roles
  rows = get_roles(accounts)
  put_rows(project.task['out']['auth'], put_json('CM_Roles', ROLES_SCHEMA),
           rows)

  # Reports
  if project.verbose:
    print('DCM Reports')

  if project.task.get('reports', False):
    rows = get_reports(accounts)
  else:
    rows = []

  put_rows(
    project.task['out']['auth'],
    put_json('CM_Reports', REPORTS_SCHEMA),
    rows
  )

  put_rows(
    project.task['out']['auth'],
    put_json('CM_Report_Deliveries', REPORT_DELIVERIES_SCHEMA),
    REPORT_DELIVERIES
  )


if __name__ == '__main__':
  barnacle()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

'''
--------------------------------------------------------------

Before running this Airflow module...

  Install StarThinker in cloud composer ( recommended ):

    From Release: pip install starthinker
    From Open Source: pip install git+https://github.com/google/starthinker

  Or push local code to the cloud composer plugins directory ( if pushing local code changes ):

    source install/deploy.sh
    4) Composer Menu
    l) Install All

--------------------------------------------------------------

  If any recipe task has "auth" set to "user" add user credentials:

    1. Ensure an RECIPE['setup']['auth']['user'] = [User Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_user", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/deploy_commandline.md#optional-setup-user-credentials

--------------------------------------------------------------

  If any recipe task has "auth" set to "service" add service credentials:

    1. Ensure an RECIPE['setup']['auth']['service'] = [Service Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_service", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/cloud_service.md

--------------------------------------------------------------

CM Log Audit

Downloads Campaign manager logs and allows audits.

  - Wait for <b>BigQuery->UNDEFINED->UNDEFINED->CM_*</b> to be created.
  - Wait for <b>BigQuery->UNDEFINED->UNDEFINED->Barnacle_*</b> to be created, then copy and connect the following data sources.
  - Join the <a hre='https://groups.google.com/d/forum/starthinker-assets' target='_blank'>StarThinker Assets Group</a> to access the following assets
  - Copy <a href='https://datastudio.google.com/open/1a6K-XdPUzCYRXZp1ZcmeOUOURc9wn2Jj' target='_blank'>Barnacle Profile Advertiser Map</a> and connect.
  - Copy <a href='https://datastudio.google.com/open/1NEzrQWWnPjkD90iUwN-ASKbVBzoeBdoT' target='_blank'>Barnacle Profile Campaign Map</a> and connect.
  - Copy <a href='https://datastudio.google.com/open/1v_GRaitwPaHHKUzfJZYNBhzotvZ-bR7Y' target='_blank'>Barnacle Profile Site Map</a> and connect.
  - Copy <a href='https://datastudio.google.com/open/14tWlh7yiqzxKJIppMFVOw2MoMtQV_ucE' target='_blank'>Barnacle Profiles Connections</a> and connect.
  - Copy <a href='https://datastudio.google.com/open/1mavjxvHSEPfJq5aW4FYgCXsBCE5rthZG' target='_blank'>Barnacle Report Delivery Profiles</a> and connect.
  - Copy <a href='https://datastudio.google.com/open/1Azk_Nul-auinf4NnDq8T9fDyiKkUWD7A' target='_blank'>Barnacle Roles Duplicates</a> and connect.
  - Copy <a href='https://datastudio.google.com/open/1ogoofpKtqkLwcW9qC_Ju_JvJdIajsjNI' target='_blank'>Barnacle Roles Not Used</a> and connect.
  - Copy <a href='https://datastudio.google.com/open/1xLgZPjOPDtmPyEqYMiMbWwMI8-WTslfj' target='_blank'>Barnacle Site Contacts Profiles</a> and connect.
  - If reports checked, copy <a href='https://datastudio.google.com/open/1-YGDiQPDnk0gD78_QOY5XdTXRlTrLeEq' target='_blank'>Barnacle Profile Report Map</a> and connect.
  - Copy <a href='https://datastudio.google.com/open/1gjxHm0jUlQUd0jMuxaOlmrl8gOX1kyKT' target='_blank'>Barnacle Report</a>.
  - When prompted choose the new data sources you just created.
  - Or give these intructions to the client.

--------------------------------------------------------------

This StarThinker DAG can be extended with any additional tasks from the following sources:
  - https://google.github.io/starthinker/
  - https://github.com/google/starthinker/tree/master/dags

'''

from starthinker.airflow.factory import DAG_Factory

INPUTS = {
  'auth_read': 'user',  # Credentials used for reading data.
  'auth_write': 'service',  # Credentials used for writing data.
  'accounts': [],  # Comma separated CM account ids.
  'days': 7,  # Number of days to backfill the log, works on first run only.
  'recipe_project': '',  # Google BigQuery project to create tables in.
  'recipe_slug': '',  # Google BigQuery dataset to create tables in.
}

RECIPE = {
  'setup': {
    'day': [
      'Mon',
      'Tue',
      'Wed',
      'Thu',
      'Fri',
      'Sat',
      'Sun'
    ],
    'hour': [
      1,
      2
    ]
  },
  'tasks': [
    {
      'dataset': {
        'description': 'The dataset will hold log table, Create it exists.',
        'hour': [
          1
        ],
        'auth': {
          'field': {
            'name': 'auth_write',
            'kind': 'authentication',
            'order': 1,
            'default': 'service',
            'description': 'Credentials used for writing data.'
          }
        },
        'dataset': {
          'field': {
            'name': 'recipe_slug',
            'kind': 'string',
            'order': 4,
            'default': '',
            'description': 'Name of Google BigQuery dataset to create.'
          }
        }
      }
    },
    {
      'dcm_log': {
        'description': 'Will create tables with format CM_* to hold each endpoint via a call to the API list function. Exclude reports for its own task.',
        'hour': [
          2
        ],
        'auth': {
          'field': {
            'name': 'auth_read',
            'kind': 'authentication',
            'order': 0,
            'default': 'user',
            'description': 'Credentials used for reading data.'
          }
        },
        'accounts': {
          'single_cell': True,
          'values': {
            'field': {
              'name': 'accounts',
              'kind': 'integer_list',
              'order': 2,
              'default': [
              ],
              'description': 'Comma separated CM account ids.'
            }
          }
        },
        'days': {
          'field': {
            'name': 'days',
            'kind': 'integer',
            'order': 3,
            'default': 7,
            'description': 'Number of days to backfill the log, works on first run only.'
          }
        },
        'out': {
          'auth': {
            'field': {
              'name': 'auth_write',
              'kind': 'authentication',
              'order': 1,
              'default': 'service',
              'description': 'Credentials used for writing data.'
            }
          },
          'project': {
            'field': {
              'name': 'recipe_project',
              'kind': 'string',
              'order': 4,
              'default': '',
              'description': 'Google BigQuery project to create tables in.'
            }
          },
          'dataset': {
            'field': {
              'name': 'recipe_slug',
              'kind': 'string',
              'order': 5,
              'default': '',
              'description': 'Google BigQuery dataset to create tables in.'
            }
          }
        }
      }
    }
  ]
}

DAG_FACTORY = DAG_Factory('dcm_log', RECIPE, INPUTS)
DAG = DAG_FACTORY.generate()

if __name__ == "__main__":
  DAG_FACTORY.print_commandline()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from django.core.management.base import BaseCommand, CommandError
from django.conf import settings

from starthinker.util.project import project
from starthinker.util.dbm import report_get as dbm_report_get
from starthinker.util.bigquery import rows_to_table, query_to_table, query_to_rows, table_exists

from starthinker_ui.account.models import Account
from starthinker_ui.recipe.models import Recipe


def account_from_dt(values):
  print('account_from_dt', values)
  return int(values['bucket'].replace('dcdt_-dcm_account', '').split('_', 1)[0])


def account_from_dbm_report(report_id, report_name):
  print('account_from_dbm_report', report_id, report_name)

  partners = []

  if report_id or report_name:
    try:
      report = dbm_report_get('user', report_id, report_name)
    except Exception as e:
      print(e, report_id, report_name)
      report = None

    if report:
      for filter in report['params']['filters']:
        if filter['type'] == 'FILTER_PARTNER':
          partners.append(filter['value'])

  return partners


class Command(BaseCommand):
  help = 'Prints recipe acounts.'

  def add_arguments(self, parser):
    parser.add_argument(
        '--recipes',
        action='store',
        dest='recipes',
        type=int,
        default=3,
        help='Number of recipes to pull at one time.',
    )

    parser.add_argument(
        '--test',
        action='store_true',
        dest='test',
        default=False,
        help='Test print rows instead of writing them.',
    )

  def handle(self, *args, **kwargs):

    impact = [
    ]  #{ 'day': DATE, 'deployment':INT, 'account': INT, 'product': STRING, 'recipe': STRING, 'user': STRING }
    missing = {}
    id_max = 0

    project.initialize(_service=settings.UI_SERVICE, _verbose=True)

    if table_exists('service', 'google.com:starthinker', 'dashboard',
                    'ST_Scripts'):
      id_max = next(
          query_to_rows(
              'service',
              'google.com:starthinker',
              'dashboard',
              'SELECT MAX(Deployment) FROM ST_Scripts',
              legacy=False))[0]

    for recipe in Recipe.objects.filter(
        id__gt=id_max).order_by('id')[:kwargs['recipes']]:

      project.initialize(
          _user=recipe.account.get_credentials_path(),
          _service=settings.UI_SERVICE,
          _verbose=True)

      values = recipe.get_values()
      for v in values:
        if v['tag'] in ('dcm_to_bigquery', 'dcm_to_sheets', 'dcm_to_storage',
                        'dcm_run', 'conversion_upload_from_bigquery',
                        'conversion_upload_from_sheets'):
          impact.append({
              'day': recipe.birthday,
              'deployment': recipe.id,
              'account': v['values'].get('account'),
              'script': v['tag'],
              'product': 'dcm',
              'user': recipe.account.email.replace('@google.com', '')
          })
        elif v['tag'] in ('dbm_to_bigquery', 'dbm_to_sheets', 'dbm_to_storage'):
          for partner in account_from_dbm_report(
              v['values'].get('dbm_report_id'),
              v['values'].get('dbm_report_name')):
            impact.append({
                'day': recipe.birthday,
                'deployment': recipe.id,
                'account': partner,
                'script': v['tag'],
                'product': 'dbm',
                'user': recipe.account.email.replace('@google.com', '')
            })
        elif v['tag'] in ('dt',):
          impact.append({
              'day': recipe.birthday,
              'deployment': recipe.id,
              'account': account_from_dt(v['values']),
              'script': v['tag'],
              'product': 'dcm',
              'user': recipe.account.email.replace('@google.com', '')
          })
        elif v['tag'] == 'barnacle':
          for account in v['values']['accounts']:
            impact.append({
                'day': recipe.birthday,
                'deployment': recipe.id,
                'account': account,
                'script': v['tag'],
                'product': 'dcm',
                'user': recipe.account.email.replace('@google.com', '')
            })
        elif v['tag'] in ('entity',):
          for partner in v['values']['partners']:
            impact.append({
                'day': recipe.birthday,
                'deployment': recipe.id,
                'account': partner,
                'script': v['tag'],
                'product': 'dbm',
                'user': recipe.account.email.replace('@google.com', '')
            })
        elif v['tag'] == 'itp':
          impact.append({
              'day': recipe.birthday,
              'deployment': recipe.id,
              'account': v['values']['dcm_account'],
              'script': v['tag'],
              'product': 'dcm',
              'user': recipe.account.email.replace('@google.com', '')
          })
          impact.append({
              'day': recipe.birthday,
              'deployment': recipe.id,
              'account': v['values']['dbm_partner'],
              'script': v['tag'],
              'product': 'dbm',
              'user': recipe.account.email.replace('@google.com', '')
          })
        elif v['tag'] == 'itp_audit':
          impact.append({
              'day': recipe.birthday,
              'deployment': recipe.id,
              'account': v['values']['cm_account_id'],
              'script': v['tag'],
              'product': 'dcm',
              'user': recipe.account.email.replace('@google.com', '')
          })
          for partner in account_from_dbm_report(
              None, v['values'].get('dv360_report_name')):
            impact.append({
                'day': recipe.birthday,
                'deployment': recipe.id,
                'account': partner,
                'script': v['tag'],
                'product': 'dbm',
                'user': recipe.account.email.replace('@google.com', '')
            })
        else:
          impact.append({
              'day': recipe.birthday,
              'deployment': recipe.id,
              'account': None,
              'script': v['tag'],
              'product': None,
              'user': recipe.account.email.replace('@google.com', '')
          })
          missing.setdefault(v['tag'], 0)
          missing[v['tag']] += 1

    if impact:
      if kwargs['test']:
        print(impact)
      else:

        print('WRITING TO ST_Scripts')
        rows_to_table(
            'service',
            'google.com:starthinker',
            'dashboard',
            'ST_Scripts', [(i['day'], i['deployment'], i['user'], i['product'],
                            i['script'], i['account']) for i in impact],
            schema=[
                {
                    'mode': 'REQUIRED',
                    'name': 'Day',
                    'type': 'Date'
                },
                {
                    'mode': 'REQUIRED',
                    'name': 'Deployment',
                    'type': 'INTEGER'
                },
                {
                    'mode': 'REQUIRED',
                    'name': 'User',
                    'type': 'STRING'
                },
                {
                    'mode': 'NULLABLE',
                    'name': 'Product',
                    'type': 'STRING'
                },
                {
                    'mode': 'NULLABLE',
                    'name': 'Recipe',
                    'type': 'STRING'
                },
                {
                    'mode': 'NULLABLE',
                    'name': 'Account',
                    'type': 'INTEGER'
                },
            ],
            skip_rows=0,
            disposition='WRITE_TRUNCATE' if id_max == 0 else 'WRITE_APPEND',
            wait=True)

      print('MISSING', missing)
      print('Coverage:', (len(impact) * 100) / (len(missing) + len(impact)))
    else:
      print('No recipes newer than:', id_max)



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################
"""Processes standard read / write JSON block for dynamic loading of data.

The goal is to create standard re-usable input output blocks.  These functions
are extensible, and should include additional sources and destinations over
time.
Key benfits include:

  - Allows single point API revision for common read / write tasks.
  - Reduce risk with test case coverage for this module inherited by recipes
  using it.
  - Using standard blocks is a big time savings.

"""

import pysftp
import traceback

from starthinker.util.project import project
from starthinker.util.storage import parse_path, makedirs_safe, object_put, bucket_create
from starthinker.util.bigquery import query_to_rows, table_to_rows, rows_to_table, json_to_table, incremental_rows_to_table, query_parameters
from starthinker.util.sheets import sheets_read, sheets_write, sheets_clear
from starthinker.util.csv import rows_to_csv


def get_rows(auth, source, as_object=False, unnest=False):
  """Processes standard read JSON block for dynamic loading of data.

  Allows us to quickly pull a column or columns of data from and use it as an
  input
  into a script. For example pull a list of ids from bigquery and act on each
  one.

  - When pulling a single column specify single_cell = True. Returns list AKA
  values.
  - When pulling a multiple columns specify single_cell = False. Returns list of
  lists AKA rows.
  - Values are always given as a list ( single_cell will trigger necessary
  wrapping ).
  - Values, bigquery, sheet are optional, if multiple given result is one
  continous iterator.
  - Extensible, add a handler to define a new source ( be kind update the
  documentation json ).

  Include the following JSON in a recipe, then in the run.py handler when
  encountering that block pass it to this function and use the returned results.

    from utils.data import get_rows

    var_json = {
      "in":{
        "single_cell":[ boolean ],
        "values": [ integer list ],
        "bigquery":{
          "dataset": [ string ],
          "query": [ string ],
          "legacy":[ boolean ]
        },
        "bigquery":{
          "dataset": [ string ],
          "table": [ string ],
        },
        "sheets":{
          "sheet":[ string - full URL, suggest using share link ],
          "tab":[ string ],
          "range":[ string - A1:A notation ]
        }
      }
    }

    values = get_rows('user', var_json)

  Or you can use it directly with project singleton.

    from util.project import project
    from utils.data import get_rows

    @project.from_parameters
    def something():
      values = get_rows(project.task['auth'], project.task['in'])

    if __name__ == "__main__":
      something()

  Args:
    auth: (string) The type of authentication to use, user or service.
    source: (json) A json block resembling var_json described above.

  Returns:
    If single_cell is False: Returns a list of row values [[v1], [v2], ... ]
    If single_cell is True: Returns a list of values [v1, v2, ...]
"""

  # if handler points to list, concatenate all the values from various sources into one list
  if isinstance(source, list):
    for s in source:
      for r in get_rows(auth, s):
        yield r

  # if handler is an endpoint, fetch data
  else:
    if 'values' in source:
      if isinstance(source['values'], list):
        for value in source['values']:
          yield value
      else:
        yield source['values']

    # should be sheets, deprecate sheet over next few releases
    if 'sheet' in source:
      rows = sheets_read(
          auth,
          source['sheet']['sheet'],
          source['sheet']['tab'],
          source['sheet']['range'],
      )

      for row in rows:
        yield row[0] if unnest or source.get('single_cell', False) else row

    if 'sheets' in source:
      rows = sheets_read(
          auth,
          source['sheets']['sheet'],
          source['sheets']['tab'],
          source['sheets']['range'],
      )

      for row in rows:
        yield row[0] if unnest or source.get('single_cell', False) else row

    if 'bigquery' in source:

      rows = []
      as_object = as_object or source['bigquery'].get('as_object', False)
      unnest = unnest or source.get('single_cell', False)

      if 'table' in source['bigquery']:
        rows = table_to_rows(
            source['bigquery'].get('auth', auth),
            project.id,
            source['bigquery']['dataset'],
            source['bigquery']['table'],
            as_object=as_object or source['bigquery'].get('as_object', False))

      else:
        rows = query_to_rows(
            source['bigquery'].get('auth', auth),
            project.id,
            source['bigquery']['dataset'],
            query_parameters(source['bigquery']['query'],
                             source['bigquery'].get('parameters', {})),
            legacy=source['bigquery'].get('legacy', False),
            as_object=as_object or source['bigquery'].get('as_object', False))

      for row in rows:
        yield row[0] if not as_object and unnest else row


def put_rows(auth, destination, rows, variant=''):
  """Processes standard write JSON block for dynamic export of data.

  Allows us to quickly write the results of a script to a destination.  For
  example
  write the results of a DCM report into BigQuery.

  - Will write to multiple destinations if specified.
  - Extensible, add a handler to define a new destination ( be kind update the
  documentation json ).

  Include the following JSON in a recipe, then in the run.py handler when
  encountering that block pass it to this function and use the returned results.

    from utils.data import put_rows

    var_json = {
      "out":{
        "bigquery":{
          "dataset": [ string ],
          "table": [ string ]
          "schema": [ json - standard bigquery schema json ],
          "header": [ boolean - true if header exists in rows ]
          "disposition": [ string - same as BigQuery documentation ]
        },
        "sheets":{
          "sheet":[ string - full URL, suggest using share link ],
          "tab":[ string ],
          "range":[ string - A1:A notation ]
          "append": [ boolean - if sheet range should be appended to ]
          "delete": [ boolean - if sheet range should be cleared before writing ]
          ]
        },
        "storage":{
          "bucket": [ string ],
          "path": [ string ]
        },
        "file":[ string - full path to place to write file ]
      }
    }

    values = put_rows('user', var_json)

  Or you can use it directly with project singleton.

    from util.project import project
    from utils.data import put_rows

    @project.from_parameters
    def something():
      values = get_rows(project.task['auth'], project.task['out'])

    if __name__ == "__main__":
      something()

  Args:
    auth: (string) The type of authentication to use, user or service.
    destination: (json) A json block resembling var_json described above. rows (
      list ) The data being written as a list object. variant (string) Appended
      to destination to differentieate multiple objects

  Returns:
    If single_cell is False: Returns a list of row values [[v1], [v2], ... ]
    If single_cell is True: Returns a list of values [v1, v2, ...]
"""

  if 'bigquery' in destination:
    skip_rows = 1 if destination['bigquery'].get('header') and destination['bigquery'].get('schema') else 0

    if destination['bigquery'].get('format', 'CSV') == 'JSON':
      json_to_table(
          destination['bigquery'].get('auth', auth),
          destination['bigquery'].get('project_id', project.id),
          destination['bigquery']['dataset'],
          destination['bigquery']['table'] + variant,
          rows,
          destination['bigquery'].get('schema', []),
          destination['bigquery'].get('disposition', 'WRITE_TRUNCATE'),
      )

    elif destination['bigquery'].get('is_incremental_load', False) == True:
      incremental_rows_to_table(
          destination['bigquery'].get('auth', auth),
          destination['bigquery'].get('project_id', project.id),
          destination['bigquery']['dataset'],
          destination['bigquery']['table'] + variant,
          rows,
          destination['bigquery'].get('schema', []),
          destination['bigquery'].get('skip_rows', skip_rows),
          destination['bigquery'].get('disposition', 'WRITE_APPEND'),
          billing_project_id=project.id)

    else:

      rows_to_table(
          destination['bigquery'].get('auth', auth),
          destination['bigquery'].get('project_id', project.id),
          destination['bigquery']['dataset'],
          destination['bigquery']['table'] + variant,
          rows,
          destination['bigquery'].get('schema', []),
          destination['bigquery'].get('skip_rows', skip_rows),
          destination['bigquery'].get('disposition', 'WRITE_TRUNCATE'),
      )

  if 'sheets' in destination:
    if destination['sheets'].get('delete', False):
      sheets_clear(
          auth,
          destination['sheets']['sheet'],
          destination['sheets']['tab'] + variant,
          destination['sheets']['range'],
      )

    sheets_write(
        auth,
        destination['sheets']['sheet'],
        destination['sheets']['tab'] + variant,
        destination['sheets']['range'],
        rows,
        destination['sheets'].get('append', False),
    )

  if 'file' in destination:
    path_out, file_ext = destination['file'].rsplit('.', 1)
    file_out = path_out + variant + '.' + file_ext
    if project.verbose:
      print('SAVING', file_out)
    makedirs_safe(parse_path(file_out))
    with open(file_out, 'w') as save_file:
      save_file.write(rows_to_csv(rows).read())

  if 'storage' in destination and destination['storage'].get(
      'bucket') and destination['storage'].get('path'):
    # create the bucket
    bucket_create(auth, project.id, destination['storage']['bucket'])

    # put the file
    file_out = destination['storage']['bucket'] + ':' + destination['storage'][
        'path'] + variant
    if project.verbose:
      print('SAVING', file_out)
    object_put(auth, file_out, rows_to_csv(rows))

  if 'sftp' in destination:
    try:
      cnopts = pysftp.CnOpts()
      cnopts.hostkeys = None

      path_out, file_out = destination['sftp']['file'].rsplit('.', 1)
      file_out = path_out + variant + file_out

      sftp = pysftp.Connection(
          host=destination['sftp']['host'],
          username=destination['sftp']['username'],
          password=destination['sftp']['password'],
          port=destination['sftp']['port'],
          cnopts=cnopts)

      if '/' in file_out:
        dir_out, file_out = file_out.rsplit('/', 1)
        sftp.cwd(dir_out)

      sftp.putfo(rows_to_csv(rows), file_out)

    except e:
      print(str(e))
      traceback.print_exc()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import re
import pytz
import json
import functools
from itertools import chain
from datetime import date, datetime, timedelta

from django.db import models
from django.conf import settings

from starthinker.util.project import project
from starthinker_ui.account.models import Account, token_generate
from starthinker_ui.project.models import Project
from starthinker_ui.recipe.scripts import Script

JOB_INTERVAL_MS = float(1600)  # milliseconds
JOB_LOOKBACK_MS = 5 * JOB_INTERVAL_MS  # 8 seconds ( must guarantee to span several pings )
JOB_RECHECK_MS = 30 * 60 * 1000  # 30 minutes

RE_SLUG = re.compile(r'[^\w]')


def utc_milliseconds(utc_timestamp=None):
  if utc_timestamp is None:
    utc_timestamp = datetime.utcnow()
  utc_epoch = datetime.utcfromtimestamp(0)
  return int((utc_timestamp - utc_epoch) / timedelta(milliseconds=1))


def utc_milliseconds_to_timezone(utm, timezone):
  return utc_to_timezone(datetime.utcfromtimestamp(int(utm / 1000)), timezone)


def utc_to_timezone(utc_timestamp, timezone):
  tz = pytz.timezone(timezone)
  return tz.normalize(utc_timestamp.replace(tzinfo=pytz.utc).astimezone(tz))


def timezone_to_utc(tz_timestamp):
  return tz_timestamp.astimezone(pytz.utc).replace(tzinfo=None)


def time_ago(timestamp):
  ago = ''
  seconds = (datetime.utcnow() - timestamp).total_seconds()

  if seconds is None:
    ago = 'Unknown'
  elif seconds == 0:
    ago = 'Just Now'
  else:
    m, s = divmod(seconds, 60)
    h, m = divmod(m, 60)
    d, h = divmod(h, 60)

    if d:
      ago += '%d Days ' % d
    if h:
      ago += '%d Hours ' % h
    if m:
      ago += '%d Minutes ' % m
    if ago == '' and s:
      ago = '1 Minute Ago'
    else:
      ago += 'Ago'

  return ago


def reference_default():
  return token_generate(Recipe, 'token', 32)


class Recipe(models.Model):
  account = models.ForeignKey(Account, on_delete=models.PROTECT, null=True)
  token = models.CharField(max_length=8, unique=True)
  reference = models.CharField(
      max_length=32, unique=True, default=reference_default)

  project = models.ForeignKey(
      Project, on_delete=models.SET_NULL, null=True, blank=True)

  name = models.CharField(max_length=64)
  active = models.BooleanField(default=True)
  manual = models.BooleanField(default=False)

  week = models.CharField(
      max_length=64,
      default=json.dumps(['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']))
  hour = models.CharField(max_length=128, default=json.dumps([3]))

  timezone = models.CharField(
      max_length=32, blank=True, default='America/Los_Angeles')

  tasks = models.TextField()

  job_utm = models.BigIntegerField(blank=True, default=0)
  job_status = models.TextField(default='{}')
  worker_uid = models.CharField(max_length=128, default='')
  worker_utm = models.BigIntegerField(blank=True, default=0)

  birthday = models.DateField(auto_now_add=True)

  _cache_log = None

  def __str__(self):
    return self.name

  def __unicode__(self):
    return self.name

  def slug(self):
    return RE_SLUG.sub('_', self.name)

  def save(self, *args, **kwargs):
    self.get_token()
    self.get_reference()
    super(Recipe, self).save(*args, **kwargs)
    self._cache_log = None

  def uid(self):
    return self.pk or 'NEW'

  def link_edit(self):
    return '/recipe/edit/%d/' % self.pk

  def link_delete(self):
    return '/recipe/delete/%d/' % self.pk

  def link_run(self):
    return '/recipe/run/%d/' % self.pk if self.pk else ''

  def link_cancel(self):
    return '/recipe/cancel/%d/' % self.pk if self.pk else ''

  def link_json(self):
    return '/recipe/json/%d/' % self.pk if self.pk else ''

  def link_colab(self):
    return '/recipe/colabs/%d/' % self.pk if self.pk else ''

  def link_airflow(self):
    return '/recipe/airflow/%d/' % self.pk if self.pk else ''

  def link_start(self):
    return '%s/recipe/start/' % settings.CONST_URL

  def link_stop(self):
    return '%s/recipe/stop/' % settings.CONST_URL

  def is_running(self):
    return self.get_log()['status'] == 'RUNNING'

  def get_token(self):
    if not self.token:
      self.token = token_generate(Recipe, 'token')
    return self.token

  def get_reference(self):
    if not self.reference:
      self.reference = token_generate(Recipe, 'reference', 32)
    return self.reference

  def get_values(self):
    constants = {
        'recipe_project':
            self.get_project_identifier(),
        'recipe_name':
            self.name,
        'recipe_slug':
            self.slug(),
        'recipe_token':
            self.get_token(),
        'recipe_timezone':
            self.timezone,
        'recipe_email':
            self.account.email if self.account else None,
        'recipe_email_token':
            self.account.email.replace('@', '+%s@' % self.get_token())
            if self.account else None,
    }
    tasks = json.loads(self.tasks or '[]')
    for task in tasks:
      task['values'].update(constants)
    return tasks

  def set_values(self, scripts):
    self.tasks = json.dumps(scripts)

  def get_hours(self):
    return [int(h) for h in json.loads(self.hour or '[]')]

  def get_days(self):
    return json.loads(
        self.week or '[]') or ['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']

  def get_icon(self):
    return ''  #get_icon('')

  def get_credentials_user(self):
    return self.account.get_credentials_path() if self.account else '{}'

  def get_credentials_service(self):
    return self.project.service if self.project and self.project.service else '{}'

  def get_project_identifier(self):
    return self.project.get_project_id() if self.project else ''

  def get_project_key(self):
    return self.project.key if self.project else ''

  def get_scripts(self):
    for value in self.get_values():
      yield Script(value['tag'])

  def get_tasks(self):
    for task in chain.from_iterable(
        map(lambda s: s.get_tasks(), self.get_scripts())):
      yield next(iter(task.items()))  # script, task

  def get_json(self, credentials=True):
    return Script.get_json(
      self.uid(),
      self.get_project_identifier(),
      self.get_project_key() if credentials else '',
      self.get_credentials_user() if credentials else '',
      self.get_credentials_service() if credentials else '', self.timezone,
      self.get_days(), self.get_hours(), self.get_values()
    )

  def activate(self):
    self.active = True
    self.save(update_fields=['active'])

  def deactivate(self):
    self.active = False
    self.save(update_fields=['active'])

  def update(self):
    return self.get_status(update=True)

  def force(self):
    status = self.get_status(force=True)
    self.worker_uid = ''  # forces current worker to cancel job
    self.save(update_fields=['worker_uid'])
    return status

  def cancel(self):
    status = self.get_status(cancel=True)
    self.worker_uid = ''  # forces current worker to cancel job
    self.save(update_fields=['worker_uid'])
    return status

  def get_job_utm(self, status):
    now_tz = utc_to_timezone(datetime.utcnow(), self.timezone)

    # check if tasks remain for today
    hour = None
    if now_tz.strftime('%a') in self.get_days():
      for task in status['tasks']:
        if task['done']:
          continue
        else:
          hour = task['hour']
          break

    # all tasks done, advance to next day first task
    if hour is None:
      now_tz += timedelta(hours=24)
      for i in range(0, 7):
        if now_tz.strftime('%a') in self.get_days():
          break
        else:
          now_tz += timedelta(hours=24)

      # get the first hour ( if tasks exist, lame use of for loop but works )
      for script, task in self.get_tasks():
        try:
          hour = task.get('hour', self.get_hours())[0]
        except IndexError:
          hour = 0
        break

    now_tz = now_tz.replace(hour=hour or 0, minute=0, second=0, microsecond=0)
    return utc_milliseconds(timezone_to_utc(now_tz))

  def get_status(self, update=False, force=False, cancel=False):
    # current 24 hour time zone derived frame to RUN the job
    now_utc = datetime.utcnow()
    now_tz = utc_to_timezone(now_utc, self.timezone)
    date_tz = str(now_tz.date())

    # load prior status
    try:
      status = json.loads(self.job_status)
    except ValueError:
      status = {}

    # create default status for new recipes
    status.setdefault('date_tz', date_tz)
    status.setdefault('tasks', [])

    # if not saved yet, do nothing
    if not self.pk:
      return status

    # if cancel, do it on whatever status exists
    elif cancel:
      for task in status['tasks']:
        if not task['done']:
          task['done'] = True
          task['utc'] = str(now_utc)
          task['event'] = 'JOB_CANCEL'

      self.job_utm = self.get_job_utm(status)
      self.job_status = json.dumps(status)
      self.worker_uid = ''  # forces current worker to cancel job
      self.save(update_fields=['job_status', 'job_utm', 'worker_uid'])

    # if manual and all task are done set the utm to be ignored in worker pulls
    elif self.manual and not force and not update:
      if not status['tasks'] or all(task['done'] for task in status['tasks']):
        self.job_utm = 0
        self.save(update_fields=['job_utm'])

    # if updating, modify the status
    elif force or update or (date_tz > status['date_tz'] and
                             now_tz.strftime('%a') in self.get_days()):
      status = {
          'date_tz': date_tz,
          'tasks': [],
      }

      # create task list based on recipe json
      instances = {}
      for order, (script, task) in enumerate(self.get_tasks()):

        # if force use current hour, if schedule use task and recipe hours
        hours = [now_tz.hour] if force else task.get('hour', self.get_hours())

        # tasks with hours = [] will be skipped unless force=True
        if hours:
          instances.setdefault(script, 0)
          instances[script] += 1
          for hour in hours:
            status['tasks'].append({
                'order': order,
                'script': script,
                'instance': instances[script],
                'hour': hour,
                'utc': str(datetime.utcnow()),
                'event': 'JOB_NEW' if update else 'JOB_PENDING',
                'stdout': '',
                'stderr': '',
                'done':
                    update  # if saved by user, write as done for that day, user must force run first time
            })

      # sort new order by first by hour and second by order
      def queue_compare(left, right):
        if left['hour'] < right['hour']:
          return -1
        elif left['hour'] > right['hour']:
          return 1
        else:
          if left['order'] < right['order']:
            return -1
          elif left['order'] > right['order']:
            return 1
          else:
            return 0

      status['tasks'].sort(key=functools.cmp_to_key(queue_compare))

      self.job_utm = self.get_job_utm(status)
      self.job_status = json.dumps(status)
      if force or update:
        self.worker_uid = ''  # cancel all current workers
        self.save(update_fields=['job_status', 'job_utm', 'worker_uid'])
      else:
        self.save(update_fields=['job_status', 'job_utm'])

    else:
      job_utm = self.get_job_utm(status)
      if job_utm != self.job_utm:
        self.job_utm = job_utm
        self.save(update_fields=['job_utm'])

    return status

  def get_task(self):
    status = self.get_status()

    # if not done return next task prior or equal to current time zone hour
    now_tz = utc_to_timezone(datetime.utcnow(), self.timezone)
    if now_tz.strftime('%a') in self.get_days():
      for task in status['tasks']:
        if not task['done'] and task['hour'] <= now_tz.hour:
          task['recipe'] = self.get_json()
          return task

    return None

  def set_task(self, script, instance, hour, event, stdout, stderr):
    status = self.get_status()

    for task in status['tasks']:
      if task['script'] == script and task['instance'] == instance and task[
          'hour'] == hour:
        task['utc'] = str(datetime.utcnow())
        task['event'] = event
        if stdout:
          task['stdout'] += stdout
        if stderr:
          task['stderr'] += stderr
        task['done'] = (event != 'JOB_START')

        self.job_status = json.dumps(status)
        self.job_utm = self.get_job_utm(status)
        self.worker_utm = utc_milliseconds(
        )  # give worker some time to clean up
        self.save(update_fields=['worker_utm', 'job_utm', 'job_status'])
        break

  def get_log(self):
    if self._cache_log is None:
      self._cache_log = self.get_status()

      error = False
      timeout = False
      new = False
      done = 0
      for task in self._cache_log['tasks']:
        task['utc'] = datetime.strptime(task['utc'].split('.', 1)[0],
                                        '%Y-%m-%d %H:%M:%S')
        task['ltc'] = utc_to_timezone(task['utc'], self.timezone)
        task['ago'] = time_ago(task['utc'])

        if task['done'] and task['event'] != 'JOB_NEW':
          done += 1
        if self._cache_log.get('utc', task['utc']) <= task['utc']:
          self._cache_log['utc'] = task['utc']

        if task['event'] == 'JOB_TIMEOUT':
          timeout = True
        elif task['event'] == 'JOB_NEW':
          new = True
        elif task['event'] not in ('JOB_PENDING', 'JOB_START', 'JOB_END'):
          error = True

      if 'utc' not in self._cache_log:
        self._cache_log['utc'] = datetime.utcnow()
      self._cache_log['utl'] = utc_to_timezone(self._cache_log['utc'],
                                               self.timezone)
      self._cache_log['ago'] = time_ago(self._cache_log['utc'])
      self._cache_log['percent'] = int(
          (done * 100) / (len(self._cache_log['tasks']) or 1))
      self._cache_log['uid'] = self.uid()

      if timeout:
        self._cache_log['status'] = 'TIMEOUT'
      elif new:
        self._cache_log['status'] = 'NEW'
      elif error:
        self._cache_log['status'] = 'ERROR'
      elif not self._cache_log['tasks'] or all(
          task['done'] for task in self._cache_log['tasks']):
        self._cache_log['status'] = 'FINISHED'
      elif utc_milliseconds() - self.worker_utm < JOB_LOOKBACK_MS:
        self._cache_log['status'] = 'RUNNING'
      elif not self.active:
        self._cache_log['status'] = 'PAUSED'
      else:
        self._cache_log['status'] = 'QUEUED'

    return self._cache_log
